from typing import Generator, Optional

from fastapi import Request, Security
from fastapi.security import OAuth2PasswordBearer
from fastapi.security.api_key import APIKeyCookie, APIKeyHeader, APIKeyQuery

from contaxy import config
from contaxy.managers.components import ComponentManager
from contaxy.schema.exceptions import UnauthenticatedError


class APITokenExtractor:
    def __init__(self, *, auto_error: bool = True):
        self.auto_error = auto_error

    async def __call__(
        self,
        api_token_query: str = Security(
            APIKeyQuery(name=config.API_TOKEN_NAME, auto_error=False)
        ),
        api_token_header: str = Security(
            APIKeyHeader(name=config.API_TOKEN_NAME, auto_error=False)
        ),
        bearer_token: str = Security(
            OAuth2PasswordBearer(tokenUrl="auth/oauth/token", auto_error=False)
        ),
        api_token_cookie: str = Security(
            APIKeyCookie(name=config.API_TOKEN_NAME, auto_error=False)
        ),
    ) -> Optional[str]:
        # TODO: already check token validity here?
        if api_token_query:
            return api_token_query
        elif api_token_header:
            return api_token_header
        elif bearer_token:
            # TODO: move the bearer token under the cookie?
            return bearer_token
        elif api_token_cookie:
            return api_token_cookie
        else:
            if self.auto_error:
                raise UnauthenticatedError("No API token was provided.")
            else:
                return None


get_api_token = APITokenExtractor(auto_error=True)
get_optional_api_token = APITokenExtractor(auto_error=False)


def get_component_manager(request: Request) -> Generator[ComponentManager, None, None]:
    """Returns the initialized component manager.

    This is used as FastAPI dependency and called for every request.
    """
    with ComponentManager(request) as component_manager:
        yield component_manager



from datetime import datetime
from enum import Enum
from typing import List, Optional

from fastapi import HTTPException, Path, status
from pydantic import BaseModel, EmailStr, Field

from contaxy.schema.exceptions import ClientValueError
from contaxy.schema.shared import MAX_DESCRIPTION_LENGTH
from contaxy.utils.fastapi_utils import as_form

USERS_KIND = "users"
ADMIN_ROLE = "roles/admin"
USER_ROLE = "roles/user"


class AccessLevel(str, Enum):
    # Map to: select, insert, update, delete

    READ = "read"  # Viewer, view: Allows admin access , Can only view existing resources. Permissions for read-only actions that do not affect state, such as viewing (but not modifying) existing resources or data.
    WRITE = "write"  # Editor, edit, Contributor : Allows read/write access , Can create and manage all types of resources but can’t grant access to others.  All viewer permissions, plus permissions for actions that modify state, such as changing existing resources.
    ADMIN = "admin"  # Owner : Allows read-only access. Has full access to all resources including the right to edit IAM, invite users, edit roles. All editor permissions and permissions for the following actions
    # UNKNOWN = "unknown"  # Deny?

    @classmethod
    def load(cls, access_level: str) -> "AccessLevel":
        try:
            return cls(access_level.strip().lower())
        except ValueError:
            raise ClientValueError(f"Access level is unknown {access_level}")
            # return cls.UNKNOWN


class TokenPurpose(str, Enum):
    USER_API_TOKEN = "user-api-token"
    PROJECT_API_TOKEN = "project-api-token"
    SERVICE_ACCESS_TOKEN = "service-access-token"
    LOGIN_TOKEN = "login-token"
    REFRESH_TOKEN = "refresh-token"  # For user sessions
    # DEPLOYMENT_TOKEN = "deployment-token"


class TokenType(str, Enum):
    SESSION_TOKEN = "session-token"
    API_TOKEN = "api-token"


class AccessToken(BaseModel):
    token: str = Field(
        ...,
        example="f4528e540a133dd53ba6809e74e16774ebe4777a",
        description="API Token.",
    )
    token_type: TokenType = Field(
        ...,
        example=TokenType.API_TOKEN,
        description="The type of the token.",
    )
    subject: str = Field(
        ...,
        example="users/dklqmomr2c8dx9cpb202dsqku",
        description="Identifies the principal that is the subject of the token. Usually refers to the user to which the token is issued to.",
    )
    scopes: List[str] = Field(
        ...,
        example=["projects#read"],
        description="List of scopes associated with the token.",
    )
    created_at: Optional[datetime] = Field(
        None,
        description="Creation date of the token.",
    )
    expires_at: Optional[datetime] = Field(
        None,
        example="2021-04-25T10:20:30.400+02:30",
        description="Date at which the token expires and, thereby, gets revoked.",
    )


class ApiToken(AccessToken):
    description: Optional[str] = Field(
        None,
        example="Token used for accesing project resources on my local machine.",
        max_length=MAX_DESCRIPTION_LENGTH,
        description="Short description about the context and usage of the token.",
    )
    created_by: Optional[str] = Field(
        None,
        example="16fd2706-8baf-433b-82eb-8c7fada847da",
        description="ID of the user that created this token.",
    )
    token_purpose: Optional[TokenPurpose] = Field(
        None,
        example=TokenPurpose.REFRESH_TOKEN,
        description="The purpose of the token.",
    )


class AuthorizedAccess(BaseModel):
    authorized_subject: str
    resource_name: Optional[str] = None
    access_level: Optional[AccessLevel] = None
    access_token: Optional[AccessToken] = None


# Oauth Specific Code
class OAuth2TokenGrantTypes(str, Enum):
    PASSWORD = "password"
    REFRESH_TOKEN = "refresh_token"
    CLIENT_CREDENTIALS = "client_credentials"
    AUTHORIZATION_CODE = "authorization_code"


# TODO: Replaced with pydantic class
# class OAuth2TokenRequestForm:
#     """OAuth2 Token Endpoint Request Form."""

#     def __init__(
#         self,
#         grant_type: OAuth2TokenGrantTypes = Form(
#             ...,
#             description="Grant type. Determines the mechanism used to authorize the creation of the tokens.",
#         ),
#         username: Optional[str] = Form(
#             None, description="Required for `password` grant type. The user’s username."
#         ),
#         password: Optional[str] = Form(
#             None, description="Required for `password` grant type. The user’s password."
#         ),
#         scope: Optional[str] = Form(
#             None,
#             description="Scopes that the client wants to be included in the access token. List of space-delimited, case-sensitive strings",
#         ),
#         client_id: Optional[str] = Form(
#             None,
#             description="The client identifier issued to the client during the registration process",
#         ),
#         client_secret: Optional[str] = Form(
#             None,
#             description=" The client secret. The client MAY omit the parameter if the client secret is an empty string.",
#         ),
#         code: Optional[str] = Form(
#             None,
#             description="Required for `authorization_code` grant type. The value is what was returned from the authorization endpoint.",
#         ),
#         redirect_uri: Optional[str] = Form(
#             None,
#             description="Required for `authorization_code` grant type. Specifies the callback location where the authorization was sent. This value must match the `redirect_uri` used to generate the original authorization_code.",
#         ),
#         refresh_token: Optional[str] = Form(
#             None,
#             description="Required for `refresh_token` grant type. The refresh token previously issued to the client.",
#         ),
#         state: Optional[str] = Form(
#             None,
#             description="An opaque value used by the client to maintain state between the request and callback. The parameter SHOULD be used for preventing cross-site request forgery.",
#         ),
#         set_as_cookie: Optional[bool] = Form(
#             False,
#             description="If `true`, the access (and refresh) token will be set as cookie instead of the response body.",
#         ),
#     ):
#         self.grant_type = grant_type
#         self.username = username
#         self.password = password
#         self.scopes = []
#         if scope:
#             self.scopes = str(scope).split()
#         self.client_id = client_id
#         self.client_secret = client_secret
#         self.code = code
#         self.redirect_uri = redirect_uri
#         self.refresh_token = refresh_token
#         self.state = state
#         self.set_as_cookie = set_as_cookie


@as_form
class OAuth2TokenRequestFormNew(BaseModel):
    """OAuth2 Token Endpoint Request Form."""

    grant_type: OAuth2TokenGrantTypes = Field(
        ...,
        description="Grant type. Determines the mechanism used to authorize the creation of the tokens.",
    )
    username: Optional[str] = Field(
        None, description="Required for `password` grant type. The user’s username."
    )
    password: Optional[str] = Field(
        None, description="Required for `password` grant type. The user’s password."
    )
    scope: Optional[str] = Field(
        None,
        description="Scopes that the client wants to be included in the access token. List of space-delimited, case-sensitive strings",
    )
    client_id: Optional[str] = Field(
        None,
        description="The client identifier issued to the client during the registration process",
    )
    client_secret: Optional[str] = Field(
        None,
        description=" The client secret. The client MAY omit the parameter if the client secret is an empty string.",
    )
    code: Optional[str] = Field(
        None,
        description="Required for `authorization_code` grant type. The value is what was returned from the authorization endpoint.",
    )
    redirect_uri: Optional[str] = Field(
        None,
        description="Required for `authorization_code` grant type. Specifies the callback location where the authorization was sent. This value must match the `redirect_uri` used to generate the original authorization_code.",
    )
    refresh_token: Optional[str] = Field(
        None,
        description="Required for `refresh_token` grant type. The refresh token previously issued to the client.",
    )
    state: Optional[str] = Field(
        None,
        description="An opaque value used by the client to maintain state between the request and callback. The parameter SHOULD be used for preventing cross-site request forgery.",
    )
    set_as_cookie: Optional[bool] = Field(
        False,
        description="If `true`, the access (and refresh) token will be set as cookie instead of the response body.",
    )


class OAuthToken(BaseModel):
    token_type: str = Field(
        ..., description="The type of token this is, typically just the string `bearer`"
    )
    access_token: str = Field(..., description="The access token string.")
    expires_in: Optional[int] = Field(
        None,
        description="The expiration time of the access token in seconds.",
    )
    refresh_token: Optional[str] = Field(
        None, description="API token to refresh the sesion token (if it expires)."
    )
    scope: Optional[str] = Field(
        None, description="The scopes contained in the access token."
    )
    id_token: Optional[str] = Field(
        None,
        description="OpenID Connect ID Token that encodes the user’s authentication information.",
    )


class OAuthTokenIntrospection(BaseModel):
    active: bool = Field(
        ...,
        description="Indicator of whether or not the presented token is currently active.",
    )
    scope: Optional[str] = Field(
        None,
        description="A space-delimited list of scopes.",
    )
    client_id: Optional[str] = Field(
        None,
        description="Client identifier for the OAuth 2.0 client that requested this token.",
    )
    username: Optional[str] = Field(
        None,
        description="Human-readable identifier for the resource owner who authorized this token.",
    )
    token_type: Optional[str] = Field(
        None,
        description="Type of the token as defined in Section 5.1 of OAuth 2.0 [RFC6749].",
    )
    exp: Optional[int] = Field(
        None,
        description="Timestamp, measured in the number of seconds since January 1 1970 UTC, indicating when this token will expire, as defined in JWT [RFC7519].",
    )
    iat: Optional[int] = Field(
        None,
        description="Timestamp, measured in the number of seconds since January 1 1970 UTC, indicating when this token was originally issued, as defined in JWT [RFC7519].",
    )
    nbf: Optional[int] = Field(
        None,
        description="Timestamp, measured in the number of seconds since January 1 1970 UTC, indicating when this token is not to be used before, as defined in JWT [RFC7519].",
    )
    sub: Optional[str] = Field(
        None,
        description="Subject of the token, as defined in JWT [RFC7519]. Usually a machine-readable identifier of the resource owner who authorized this token.",
    )
    aud: Optional[str] = Field(
        None,
        description="Service-specific string identifier or list of string identifiers representing the intended audience for this token, as defined in JWT [RFC7519].",
    )
    iss: Optional[str] = Field(
        None,
        description="String representing the issuer of this token, as defined in JWT [RFC7519].",
    )
    jti: Optional[str] = Field(
        None,
        description="String identifier for the token, as defined in JWT [RFC7519].",
    )
    uid: Optional[str] = Field(
        None,
        description="The user ID. This parameter is returned only if the token is an access token and the subject is an end user.",
    )


class AuthorizeResponseType(str, Enum):
    TOKEN = "token"
    CODE = "code"


class OAuth2ErrorDetails(BaseModel):
    error: str


class OAuth2Error(HTTPException):
    """Basic exception for OAuth errors.

    Implements the [RFC6749 error response](https://tools.ietf.org/html/rfc6749#section-5.2).
    """

    def __init__(
        self,
        error: str,
    ) -> None:
        """Initializes the exception.

        Args:
            error: A single ASCII error code from the ones defined in RFC6749.
        """

        super(OAuth2Error, self).__init__(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=error,
        )


# TODO: Not used right now
# class OAuth2AuthorizeRequestForm:
#     """OAuth2 Authorize Endpoint Request Form."""

#     def __init__(
#         self,
#         response_type: AuthorizeResponseType = Form(
#             ...,
#             description="Either code for requesting an authorization code or token for requesting an access token (implicit grant).",
#         ),
#         client_id: Optional[str] = Form(
#             None, description="The public identifier of the client."
#         ),
#         redirect_uri: Optional[str] = Form(None, description="Redirection URL."),
#         scope: Optional[str] = Form(
#             None, description="The scope of the access request."
#         ),
#         state: Optional[str] = Form(
#             None,
#             description="An opaque value used by the client to maintain state between the request and callback. The parameter SHOULD be used for preventing cross-site request forgery",
#         ),
#         nonce: Optional[str] = Form(None),
#     ):
#         self.response_type = response_type
#         self.client_id = client_id
#         self.redirect_uri = redirect_uri
#         self.scope = scope
#         self.state = state
#         self.nonce = nonce


USER_ID_PARAM = Path(
    ...,
    title="User ID",
    description="A valid user ID.",
    # TODO: add length restriction
)

# User Models


class UserBase(BaseModel):
    username: Optional[str] = Field(
        None,
        example="john-doe",
        description="A unique username on the system.",
    )  # nickname
    email: Optional[EmailStr] = Field(
        None, example="john.doe@example.com", description="User email address."
    )
    disabled: Optional[bool] = Field(
        False,
        description="Indicates that user is disabled. Disabling a user will prevent any access to user-accesible resources.",
    )


class UserInput(UserBase):
    pass


class UserRegistration(UserInput):
    # The password is only part of the user input object and should never returned
    # TODO: a password can only be changed when used via oauth password bearer
    # TODO: System admin can change passwords for all users
    password: Optional[str] = Field(
        None,
        description="Password for the user. The password will be stored in as a hash.",
    )


class User(UserBase):
    id: str = Field(
        None,
        example="16fd2706-8baf-433b-82eb-8c7fada847da",
        description="Unique ID of the user.",
    )
    technical_user: Optional[bool] = Field(
        False,
        description="Indicates if the user is a technical user created by the system.",
    )
    created_at: Optional[datetime] = Field(
        None,
        description="Timestamp of the user creation. Assigned by the server and read-only.",
    )



from abc import ABC, abstractmethod
from typing import List

from contaxy.schema.system import AllowedImageInfo, SystemInfo, SystemStatistics


class SystemOperations(ABC):
    @abstractmethod
    def get_system_info(self) -> SystemInfo:
        pass

    @abstractmethod
    def is_healthy(self) -> bool:
        pass

    @abstractmethod
    def get_system_statistics(self) -> SystemStatistics:
        pass

    @abstractmethod
    def initialize_system(self) -> None:
        pass

    @abstractmethod
    def list_allowed_images(self) -> List[AllowedImageInfo]:
        pass

    @abstractmethod
    def add_allowed_image(self, allowed_image: AllowedImageInfo) -> AllowedImageInfo:
        pass

    @abstractmethod
    def delete_allowed_image(self, allowed_image_name: str) -> None:
        pass



"""Command line interface."""

from pathlib import Path

import typer
from loguru import logger

cli = typer.Typer()


@cli.command()
def install_contaxy() -> None:
    from contaxy.managers.components import install_components

    install_components()


@cli.command()
def export_openapi_specs(output_path: Path) -> None:
    import json

    from contaxy.api import app

    # write openapi.json spec to file
    with open(output_path, "w") as outfile:
        json.dump(app.openapi(), outfile)


@cli.command()
def start_dev_server(host: str = "0.0.0.0", port: int = 8082) -> None:
    import sys

    import uvicorn

    logger.add(
        sys.stderr, format="{time} {level} {message}", filter="my_module", level="INFO"
    )

    typer.echo("Starting Dev Server.")

    # TODO: Run with: Run with: PYTHONASYNCIODEBUG=1
    uvicorn.run("contaxy.api:app", host=host, port=port, log_level="info", reload=True)


if __name__ == "__main__":
    cli()



from typing import List

from starlette.responses import Response

from contaxy.config import settings
from contaxy.operations import DeploymentOperations
from contaxy.schema import JobInput, ResourceAction, ServiceInput

ACTION_DELIMITER = "-"
ACTION_ACCESS = "access"


class DeploymentManager(DeploymentOperations):
    def execute_service_action(
        self,
        project_id: str,
        service_id: str,
        action_id: str,
    ) -> Response:
        # TODO: redirect from web app fetch / request does not work as the request itself is redirected and not the page.
        # try:
        #     if action_id.startswith(ACTION_ACCESS):
        #         port = action_id.split(ACTION_DELIMITER)[1]

        #         return RedirectResponse(
        #             url=f"{settings.CONTAXY_BASE_URL}/projects/{project_id}/services/{service_id}/access/{port}"
        #         )
        # except Exception:
        #     raise ServerBaseError("Could not execute action")

        # return Response(
        #     content=f"No implementation for action id '{action_id}'", status_code=501
        # )
        raise NotImplementedError

    def execute_job_action(
        self, project_id: str, job_id: str, action_id: str
    ) -> Response:
        # 501: not implemented
        return Response(status_code=501)

    def list_service_actions(
        self, project_id: str, service_id: str
    ) -> List[ResourceAction]:
        service_metadata = self.get_service_metadata(
            project_id=project_id, service_id=service_id
        )

        resource_actions: List[ResourceAction] = []
        if service_metadata.endpoints:
            for endpoint in service_metadata.endpoints:
                endpoint = endpoint.replace("/", "")
                resource_actions.append(
                    ResourceAction(
                        action_id=f"{ACTION_ACCESS}{ACTION_DELIMITER}{endpoint}",
                        display_name=f"Endpoint: {endpoint}",
                        instructions=[
                            {
                                "type": "new-tab",
                                "url": f"{settings.CONTAXY_BASE_URL}/projects/{project_id}/services/{service_id}/access/{endpoint}",
                            }
                        ],
                    )
                )

        return resource_actions

    def list_job_actions(
        self,
        project_id: str,
        job_id: str,
    ) -> List[ResourceAction]:
        return []

    def suggest_service_config(
        self,
        project_id: str,
        container_image: str,
    ) -> ServiceInput:
        # TODO: add sensible logic here
        return ServiceInput(container_image=container_image)

    def suggest_job_config(self, project_id: str, container_image: str) -> JobInput:
        # TODO: add sensible logic here
        return JobInput(container_image=container_image)



import io
from random import randint
from typing import List, Optional

from faker import Faker
from pydantic import EmailStr

from contaxy.operations import AuthOperations, FileOperations, ProjectOperations
from contaxy.operations.seed import SeedOperations
from contaxy.schema import File, Project, User, UserRegistration
from contaxy.schema.project import ProjectCreation
from contaxy.utils.auth_utils import setup_user
from contaxy.utils.file_utils import FileStreamWrapper
from contaxy.utils.state_utils import GlobalState, RequestState

FAKER = Faker()
Faker.seed(0)

ERROR_NO_PROJECT_MANAGER = "Seeder needs to be initialized with a project manager"


class SeedManager(SeedOperations):
    def __init__(
        self,
        global_state: Optional[GlobalState] = None,
        request_state: Optional[RequestState] = None,
        auth_manager: Optional[AuthOperations] = None,
        file_manager: Optional[FileOperations] = None,
        project_manager: Optional[ProjectOperations] = None,
    ):
        self.global_state = global_state
        self.request_state = request_state
        self.auth_manager = auth_manager
        self.file_manager = file_manager
        self.project_manager = project_manager

    def create_user(
        self,
        user_input: UserRegistration = UserRegistration(
            username="Foo", email=EmailStr("foo@bar.com"), password="Foobar"
        ),
    ) -> User:
        if not self.auth_manager:
            raise RuntimeError("Seeder needs to be initialized with an auth manager")
        user = self.auth_manager.create_user(user_input=user_input)
        self.setup_user(user)
        return user

    def create_users(self, amount: int) -> List[User]:
        users = []
        for _ in range(amount):
            user = self.create_user(
                UserRegistration(
                    username=FAKER.user_name(),
                    email=FAKER.email(),
                    password=FAKER.password(),
                )
            )
            users.append(user)

        return users

    def setup_user(self, user: User) -> Project:
        if not self.project_manager:
            raise RuntimeError(ERROR_NO_PROJECT_MANAGER)
        if not self.auth_manager:
            raise RuntimeError("Seeder needs to be initialized with an auth manager")
        return setup_user(user, self.project_manager)

    def create_project(
        self,
        project_input: ProjectCreation = ProjectCreation(
            id="my-test-project", display_name="My Test Project!"
        ),
    ) -> Project:
        if not self.project_manager:
            raise RuntimeError(ERROR_NO_PROJECT_MANAGER)
        return self.project_manager.create_project(project_input=project_input)

    def create_projects(self, amount: int) -> List[Project]:
        if not self.project_manager:
            raise RuntimeError(ERROR_NO_PROJECT_MANAGER)
        projects = []
        for _ in range(amount):
            project = self.create_project(
                ProjectCreation(id=FAKER.ean(), display_name=FAKER.word())
            )
            projects.append(project)

        return projects

    def create_file(
        self,
        project_id: str,
        file_key: str = "my-test-file",
        max_number_chars: int = 200,
    ) -> File:
        if not self.file_manager:
            raise RuntimeError("Seeder needs to be initialized with a file manager")
        file_stream = FileStreamWrapper(
            io.BytesIO(FAKER.text(max_nb_chars=max_number_chars).encode("UTF-8"))
        )
        return self.file_manager.upload_file(
            project_id=project_id, file_key=file_key, file_stream=file_stream  # type: ignore
        )

    def create_files(
        self,
        project_id: str,
        number_of_files: int,
        prefix: Optional[str] = "my-test-file",
        max_number_chars: int = 200,
    ) -> List[File]:
        return [
            self.create_file(
                project_id,
                f"{prefix}-{randint(1,10000)}",
                max_number_chars,
            )
            for index in range(number_of_files)
        ]

    def create_file_stream(
        self,
        max_number_chars: int = 200,
    ) -> FileStreamWrapper:
        return FileStreamWrapper(
            io.BytesIO(FAKER.text(max_nb_chars=max_number_chars).encode("UTF-8"))
        )



from loguru import logger
from sqlalchemy.exc import ProgrammingError
from sqlalchemy.future import Engine
from sqlalchemy.schema import CreateSchema


def create_schema(engine: Engine, schema_name: str) -> None:
    with engine.begin() as conn:
        stmt = CreateSchema(schema_name)
        try:
            conn.execute(stmt)
            conn.commit()
            # logger.debug(f"Postgres DB Schema {schema_name} created")
        except ProgrammingError:
            logger.debug(
                f"Postgres DB Schema {schema_name} not created. This because it already exists."
            )



from .auth import AuthClient
from .deployment import DeploymentManagerClient
from .extension import ExtensionClient
from .file import FileClient
from .json_db import JsonDocumentClient
from .project import ProjectClient
from .system import SystemClient



from typing import Dict, List, Optional

import requests
from pydantic.tools import parse_raw_as

from contaxy.clients.shared import handle_errors
from contaxy.operations import ExtensionOperations
from contaxy.schema.extension import Extension, ExtensionInput


class ExtensionClient(ExtensionOperations):
    def __init__(self, client: requests.Session):
        self._client = client

    @property
    def client(self) -> requests.Session:
        return self._client

    def list_extensions(
        self, project_id: str, request_kwargs: Dict = {}
    ) -> List[Extension]:
        response = self.client.get(
            f"/projects/{project_id}/extensions", **request_kwargs
        )
        handle_errors(response)
        return parse_raw_as(List[Extension], response.text)

    def install_extension(
        self, extension: ExtensionInput, project_id: str, request_kwargs: Dict = {}
    ) -> Extension:
        response = self.client.post(
            f"/projects/{project_id}/extensions",
            json=extension.dict(exclude_unset=True),
            **request_kwargs,
        )
        handle_errors(response)
        return parse_raw_as(Extension, response.text)

    def delete_extension(
        self, project_id: str, extension_id: Optional[str] = None
    ) -> None:
        raise NotImplementedError

    def suggest_extension_config(
        self,
        container_image: str,
        project_id: str,
    ) -> ExtensionInput:
        raise NotImplementedError

    def get_extension_metadata(
        self,
        project_id: str,
        extension_id: str,
    ) -> Extension:
        raise NotImplementedError



from loguru import logger
from minio import Minio
from minio.deleteobjects import DeleteObject
from minio.error import S3Error
from minio.versioningconfig import ENABLED, VersioningConfig
from urllib3.exceptions import MaxRetryError

from contaxy.config import Settings
from contaxy.schema.exceptions import ServerBaseError


def get_bucket_name(project_id: str, prefix: str) -> str:
    return project_id if not prefix else f"{prefix}.{project_id}"


def create_minio_client(settings: Settings) -> Minio:
    client = Minio(
        endpoint=settings.S3_ENDPOINT,
        access_key=settings.S3_ACCESS_KEY,
        secret_key=settings.S3_SECRET_KEY,
        region=settings.S3_REGION,
        secure=settings.S3_SECURE,
    )

    try:
        if not client.bucket_exists("nonexistingbucket"):
            logger.debug("Object storage connected")
    except (MaxRetryError, S3Error):
        logger.critical(
            f"Could not connect to object storage (endpoint: {settings.S3_ENDPOINT}, region: {settings.S3_REGION}, secure: {settings.S3_SECURE})."
        )
        raise ServerBaseError(
            "Could not connect to object storage. Check S3 endpoint configuration."
        )
    return client


def purge_bucket(client: Minio, bucket_name: str) -> None:
    # TODO: Validate the effect of `bypass_governance_mode`
    logger.debug(f"Purging bucket {bucket_name}.")
    delete_objects = []
    for file in client.list_objects(bucket_name, include_version=True, recursive=True):
        delete_objects.append(DeleteObject(file.object_name, file.version_id))

    delete_count = len(delete_objects)
    if delete_objects:
        # TODO: Validate bypass_governance_mode behavior
        logger.debug(f"{delete_count} files from {bucket_name} to be removed.")
        errors = client.remove_objects(
            bucket_name, delete_objects, bypass_governance_mode=True
        )

        for error in errors:
            # TODO: Critical is only for testing an should be handled when tested
            logger.critical(
                f"Error purging bucket {bucket_name} (Errors: {error}).",
            )

        actual_deleted = delete_count - len(list(errors))
        logger.debug(
            f"{actual_deleted} files from {delete_count} from {bucket_name} deleted."
        )


def delete_bucket(client: Minio, bucket_name: str, force: bool = False) -> None:
    try:
        client.remove_bucket(bucket_name)
        logger.info(f"Bucket {bucket_name} deleted")
    except S3Error as err:
        if err.code == "NoSuchBucket":
            logger.info(f"Bucket {bucket_name} not deleted, since it does not exist.")
            return
        elif err.code == "BucketNotEmpty":
            if force:
                purge_bucket(client, bucket_name)
                client.remove_bucket(bucket_name)
                logger.info(f"Bucket {bucket_name} deleted")
            else:
                msg = (
                    f"Bucket {bucket_name} can not be deleted because it is not empty."
                )
                logger.error(msg)
                raise ServerBaseError(msg)
        else:
            raise ServerBaseError(
                f"The bucket {bucket_name} could not be deleted ({err.code})."
            )


def create_bucket(
    client: Minio,
    bucket_name: str,
    versioning_enabled: bool = True,
) -> None:
    """Create a bucket in the object storage configured by the client.

    Args:
        client (Minio): The initilaized Minio client.
        bucket_name (str): Bucket name.
        versioning_enabled (bool, optional): Controls bucket versioning. Defaults to True.

    Raises:
        ServerBaseError: If the desired action could not be performed.
    """
    try:
        client.make_bucket(bucket_name)
        logger.info(f"Bucket {bucket_name} created.")
    except S3Error as err:
        if err.code == "BucketAlreadyOwnedByYou":
            logger.info(f"The bucket {bucket_name} already exists and is owned by you.")
        else:
            logger.error(
                f"Could not create create the bucket {bucket_name} for an unknown reason."
            )
            raise ServerBaseError(f"Could not create create the bucket {bucket_name}.")

    if versioning_enabled:
        try:
            client.set_bucket_versioning(bucket_name, VersioningConfig(ENABLED))
        except S3Error as err:
            msg = ""
            if err.code == "NotImplemented":
                msg = f"Versioning could not be enabled on the bucket {bucket_name}. Ensure that the object storage is configured with versioning enabled."
            else:
                msg = f"Versioning could not be enabled on the bucket {bucket_name} for an unknown reason."

            logger.critical(msg)
            delete_bucket(client, bucket_name)
            raise ServerBaseError(msg)



import json
from typing import Any, Dict, List, Optional

from fastapi import APIRouter, Depends, Path, Query, Response, status

from contaxy.api.dependencies import (
    ComponentManager,
    get_api_token,
    get_component_manager,
)
from contaxy.schema import CoreOperations, JsonDocument
from contaxy.schema.auth import AccessLevel
from contaxy.schema.exceptions import (
    AUTH_ERROR_RESPONSES,
    CREATE_RESOURCE_RESPONSES,
    GET_RESOURCE_RESPONSES,
    UPDATE_RESOURCE_RESPONSES,
    VALIDATION_ERROR_RESPONSE,
)
from contaxy.schema.project import PROJECT_ID_PARAM

router = APIRouter(
    tags=["json"], responses={**AUTH_ERROR_RESPONSES, **VALIDATION_ERROR_RESPONSE}
)


@router.put(
    "/projects/{project_id}/json/{collection_id}/{key}",
    operation_id=CoreOperations.CREATE_JSON_DOCUMENT.value,
    summary="Create JSON document.",
    response_model=JsonDocument,
    response_model_exclude_unset=True,
    status_code=status.HTTP_200_OK,
    responses={**CREATE_RESOURCE_RESPONSES},
)
def create_json_document(
    json_document: Dict,
    project_id: str = PROJECT_ID_PARAM,
    collection_id: str = Path(..., description="ID of the collection."),
    key: str = Path(..., description="Key of the JSON document."),
    upsert: Optional[bool] = Query(
        True,
        description="If `True`, the document will be updated/overwritten if it already exists.",
    ),
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Creates a JSON document. If a document already exists for the given key, the document will be overwritten.

    If no collection exists in the project with the provided `collection_id`, a new collection will be created.
    """
    component_manager.verify_access(
        token, f"projects/{project_id}/json/{collection_id}/{key}", AccessLevel.WRITE
    )

    if upsert is None:
        # True is the default
        upsert = True

    return component_manager.get_json_db_manager().create_json_document(
        project_id, collection_id, key, json.dumps(json_document), upsert=upsert
    )


@router.patch(
    "/projects/{project_id}/json/{collection_id}/{key}",
    operation_id=CoreOperations.UPDATE_JSON_DOCUMENT.value,
    summary="Update a JSON document.",
    response_model=JsonDocument,
    response_model_exclude_unset=True,
    status_code=status.HTTP_200_OK,
    responses={**UPDATE_RESOURCE_RESPONSES},
)
def update_json_document(
    json_document: Dict,
    project_id: str = PROJECT_ID_PARAM,
    collection_id: str = Path(..., description="ID of the collection."),
    key: str = Path(..., description="Key of the JSON document."),
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Updates a JSON document.

    The update is applied on the existing document based on the JSON Merge Patch Standard [RFC7396](https://tools.ietf.org/html/rfc7396).
    """
    component_manager.verify_access(
        token, f"projects/{project_id}/json/{collection_id}/{key}", AccessLevel.WRITE
    )

    return component_manager.get_json_db_manager().update_json_document(
        project_id, collection_id, key, json.dumps(json_document)
    )


@router.get(
    "/projects/{project_id}/json/{collection_id}",
    operation_id=CoreOperations.LIST_JSON_DOCUMENTS.value,
    response_model=List[JsonDocument],
    response_model_exclude_unset=True,
    summary="List JSON documents.",
    status_code=status.HTTP_200_OK,
)
def list_json_documents(
    project_id: str = PROJECT_ID_PARAM,
    collection_id: str = Path(..., description="ID of the collection."),
    filter: Optional[str] = Query(
        None, description="JSON Path query used to filter the results."
    ),
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Lists all JSON documents for the given project collection.

    If extensions are registered for this operation and no extension is selected via the `extension_id` parameter, the results from all extensions will be included in the returned list.

    The `filter` parameter allows to filter the result documents based on a JSONPath expression ([JSON Path Specification](https://goessner.net/articles/JsonPath/)). The filter is only applied to filter documents in the list. It is not usable to extract specific properties.

    # TODO: Add filter examples
    """
    component_manager.verify_access(
        token, f"projects/{project_id}/json/{collection_id}", AccessLevel.READ
    )

    return component_manager.get_json_db_manager().list_json_documents(
        project_id, collection_id, filter
    )


@router.get(
    "/projects/{project_id}/json/{collection_id}/{key}",
    operation_id=CoreOperations.GET_JSON_DOCUMENT.value,
    response_model=JsonDocument,
    response_model_exclude_unset=True,
    summary="Get JSON document.",
    status_code=status.HTTP_200_OK,
    responses={**GET_RESOURCE_RESPONSES},
)
def get_json_document(
    project_id: str = PROJECT_ID_PARAM,
    collection_id: str = Path(..., description="ID of the collection."),
    key: str = Path(..., description="Key of the JSON document."),
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Returns a single JSON document."""

    component_manager.verify_access(
        token, f"projects/{project_id}/json/{collection_id}/{key}", AccessLevel.READ
    )

    return component_manager.get_json_db_manager().get_json_document(
        project_id, collection_id, key
    )


@router.delete(
    "/projects/{project_id}/json/{collection_id}/{key}",
    operation_id=CoreOperations.DELETE_JSON_DOCUMENT.value,
    summary="Delete JSON document.",
    status_code=status.HTTP_204_NO_CONTENT,
)
def delete_json_document(
    project_id: str = PROJECT_ID_PARAM,
    collection_id: str = Path(..., description="ID of the collection."),
    key: str = Path(..., description="Key of the JSON document."),
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Deletes a single JSON document.

    If no other document exists in the project collection, the collection will be deleted.
    """
    component_manager.verify_access(
        token, f"projects/{project_id}/json/{collection_id}/{key}", AccessLevel.WRITE
    )

    component_manager.get_json_db_manager().delete_json_document(
        project_id, collection_id, key
    )
    return Response(status_code=status.HTTP_204_NO_CONTENT)


@router.delete(
    "/projects/{project_id}/json/{collection_id}",
    operation_id=CoreOperations.DELETE_JSON_COLLECTION.value,
    summary="Delete a JSON collection.",
    status_code=status.HTTP_204_NO_CONTENT,
)
def delete_json_collection(
    project_id: str = PROJECT_ID_PARAM,
    collection_id: str = Path(..., description="ID of the collection."),
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Deletes all documents of a single JSON collection."""
    component_manager.verify_access(
        token, f"projects/{project_id}/json/{collection_id}", AccessLevel.WRITE
    )

    component_manager.get_json_db_manager().delete_json_collection(
        project_id, collection_id
    )
    return Response(status_code=status.HTTP_204_NO_CONTENT)


@router.delete(
    "/projects/{project_id}/json",
    operation_id=CoreOperations.DELETE_JSON_COLLECTIONS.value,
    summary="Delete all JSON collections.",
    status_code=status.HTTP_204_NO_CONTENT,
)
def delete_json_collections(
    project_id: str = PROJECT_ID_PARAM,
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Deletes all JSON collections for the given project."""
    component_manager.verify_access(
        token, f"projects/{project_id}/json", AccessLevel.ADMIN
    )

    component_manager.get_json_db_manager().delete_json_collections(project_id)
    return Response(status_code=status.HTTP_204_NO_CONTENT)



from datetime import datetime
from enum import Enum
from typing import Dict, List, Optional

from fastapi import Path
from pydantic import BaseModel, Field

from contaxy.schema.shared import Resource, ResourceInput

SERVICE_ID_PARAM = Path(
    ...,
    title="Service ID",
    description="A valid Service ID.",
    # TODO: add length restriction
)

JOB_ID_PARAM = Path(
    ...,
    title="Job ID",
    description="A valid job ID.",
    # TODO: add length restriction
)


class DeploymentType(str, Enum):
    CORE_BACKEND = "core-backend"
    SERVICE = "service"
    JOB = "job"
    EXTENSION = "extension"


class DeploymentStatus(str, Enum):
    # Deployment created, but not ready for usage
    PENDING = "pending"  # Alternative naming: waiting
    # Deployment is running and usable
    RUNNING = "running"
    # Deployment was stopped with successful exit code (== 0).
    SUCCEEDED = "succeeded"
    # Deployment was stopped with failure exit code (> 0).
    FAILED = "failed"
    # Deployment was deleted and is now terminating the pods.
    TERMINATING = "terminating"
    # Deployment state cannot be obtained.
    UNKNOWN = "unknown"
    # Deployment is paused (only on docker?)/
    # PAUSED = "paused"
    # Other possible options:
    # KILLED = "killed"
    # STARTING("starting"),
    # STOPPPED("stopped"),
    # CREATED("created"),
    # REBOOTING
    # TERMINATED: Container is terminated. This container can’t be started later on.
    # STOPPED:  Container is stopped. This container can be started later on.
    # ERROR – Container is an error state. Usually no operations can be performed on the container once it ends up in the error state.
    # SUSPENDED: Container is suspended.


class DeploymentCompute(BaseModel):
    min_cpus: Optional[int] = Field(
        None,
        example=2,
        ge=0,
        description="Minimum number of CPU cores required by this deployment. The system will make sure that atleast this amount is available to the deployment.",
    )
    max_cpus: Optional[int] = Field(
        None,
        example=4,
        ge=0,
        description="Maximum number of CPU cores. Even so the system will try to provide the specified amount, it's only guaranteed that the deployment cannot use more. 0 means unlimited.",
    )
    min_memory: Optional[int] = Field(
        None,
        example=4000,
        ge=5,  # 4 is the minimal RAM needed for containers
        description="Minimum amount of memory in Megabyte required by this deployment. The system will make sure that atleast this amount is available to the deployment.",
    )
    max_memory: Optional[int] = Field(
        None,
        example=8000,
        ge=0,
        description="Maximum amount of memory in Megabyte. Even so the system will try to provide the specified amount, it's only guaranteed that the deployment cannot use more. 0 means unlimited.",
    )  # in MB
    min_gpus: Optional[int] = Field(
        None,
        example=1,
        ge=0,
        description="Minimum number of GPUs required by this deployments. The system will make sure that atleast this amount is available to the deployment.",
    )
    max_gpus: Optional[int] = Field(
        None,
        example=2,
        ge=0,
        description="Maximum number of GPUs. Even so the system will try to provide the specified amount, it's only guaranteed that the deployment cannot use more.",
    )
    volume_path: Optional[str] = Field(
        None,
        example="/path/to/data",
        description="Container internal directory that should mount a volume for data persistence.",
    )
    # TODO: min_volume_size
    max_volume_size: Optional[int] = Field(
        None,
        example=32000,
        ge=1,
        description="Maximum volume size in Megabyte. This is only applied in combination with volume_path.",
    )
    # TODO: min_container_size
    max_container_size: Optional[int] = Field(
        None,
        example=32000,
        ge=1,
        description="Maximum container size in Megabyte. The deployment will be killed if it grows above this limit.",
    )
    # TODO: min_replicas
    max_replicas: Optional[int] = Field(
        1,
        example=2,
        ge=1,
        description="Maximum number of deployment instances. The system will make sure to optimize the deployment based on the available resources and requests. Use 1 if the deployment is not scalable.",
    )
    # TODO: use timedelta
    min_lifetime: Optional[int] = Field(
        None,
        example=86400,
        description="Minimum guaranteed lifetime in seconds. Once the lifetime is reached, the system is allowed to kill the deployment in case it requires additional resources.",
    )


class DeploymentBase(BaseModel):
    container_image: str = Field(
        ...,
        example="hello-world:latest",
        max_length=2000,
        description="The container image used for this deployment.",
    )
    parameters: Optional[Dict[str, str]] = Field(
        None,
        example={"TEST_PARAM": "param-value"},
        description="Parmeters (enviornment variables) for this deployment.",
    )
    compute: Optional[DeploymentCompute] = Field(
        None,
        description="Compute instructions and limitations for this deployment.",
    )
    command: Optional[str] = Field(
        None,
        description="Command to run within the deployment. This overwrites the existing entrypoint.",
    )
    requirements: Optional[List[str]] = Field(
        None,
        description="Additional requirements for deployment.",
    )
    endpoints: Optional[List[str]] = Field(
        None,
        example=["8080", "9001/webapp/ui", "9002b"],
        description="A list of HTTP endpoints that can be accessed. This should always have an internal port and can include additional instructions, such as the URL path.",
    )
    # TODO: v2
    # input_files: Optional[List[dict]] = Field(
    #    None,
    #    description="A list of files that should be added to the deployment.",
    # )
    # TODO: v2
    # command_args: Optional[List[str]] = Field(
    #    None,
    #    description="Arguments to use for the command of the deployment. This overwrites the existing arguments.",
    # )


class DeploymentInput(ResourceInput, DeploymentBase):
    pass


class Deployment(Resource, DeploymentBase):
    started_at: Optional[datetime] = Field(
        None,
        description="Timestamp when the deployment was started.",
    )
    stopped_at: Optional[datetime] = Field(
        None,
        description="Timestamp when the container has stopped.",
    )
    extension_id: Optional[str] = Field(
        None,
        description="The extension ID in case the deployment is deployed via an extension.",
    )
    deployment_type: Optional[DeploymentType] = Field(
        None,
        description="The type of this deployment.",
    )
    status: Optional[DeploymentStatus] = Field(
        None,
        example=DeploymentStatus.RUNNING,
        description="The status of this deployment.",
    )
    internal_id: Optional[str] = Field(
        None,
        example="73d247087fea5bfb3a67e98da6a07f5bf4e2a90e5b52f3c12875a35600818376",
        description="The ID of the deployment on the orchestration platform.",
    )
    # TODO: All labels should be transformed into the metadata or additional_metadata
    # deployment_labels: Optional[Dict[str, str]] = Field(
    #    None,
    #    example={"foo.bar.label": "label-value"},
    #    description="The labels of the deployment on the orchestration platform.",
    # )
    # TODO: should be a debug information.
    # exit_code: Optional[int] = Field(
    #    None,
    #    example=0,
    #    description="The Exit code of the container, in case the container was stopped.",
    # )


class ServiceBase(BaseModel):
    graphql_endpoint: Optional[str] = Field(
        None,
        example="8080/graphql",
        description="GraphQL endpoint.",
    )
    openapi_endpoint: Optional[str] = Field(
        None,
        example="8080/openapi.yaml",
        description="Endpoint that prorvides an OpenAPI schema definition..",
    )
    health_endpoint: Optional[str] = Field(
        None,
        example="8080/healthz",
        description="The endpoint instruction that can be used for checking the deployment health.",
    )


class ServiceInput(ServiceBase, DeploymentInput):
    pass


class Service(ServiceBase, Deployment):
    pass


class JobBase(BaseModel):
    pass


class JobInput(JobBase, DeploymentInput):
    pass
    # TODO: v2
    # output_files: Optional[List[dict]] = Field(
    #    None,
    #    description="A list of container internal files that should be uploaded to the storage once the job has succeeded.",
    # )


class Job(JobBase, Deployment):
    pass



from typing import Dict, List

import requests
from pydantic import parse_raw_as

from contaxy.clients.shared import handle_errors
from contaxy.operations import ProjectOperations
from contaxy.schema import AccessLevel, Project, ProjectCreation, ProjectInput, User


class ProjectClient(ProjectOperations):
    def __init__(self, client: requests.Session):
        self._client = client

    def list_projects(
        self,
        request_kwargs: Dict = {},
    ) -> List[Project]:
        response = self._client.get("/projects", **request_kwargs)
        handle_errors(response)
        return parse_raw_as(List[Project], response.text)

    def create_project(
        self,
        project_input: ProjectCreation,
        technical_project: bool = False,
        request_kwargs: Dict = {},
    ) -> Project:
        resource = self._client.post(
            "/projects",
            params={"technical_project": technical_project},
            data=project_input.json(exclude_unset=True),
            **request_kwargs,
        )
        handle_errors(resource)
        return parse_raw_as(Project, resource.text)

    def get_project(self, project_id: str, request_kwargs: Dict = {}) -> Project:
        resource = self._client.get(f"/projects/{project_id}", **request_kwargs)
        handle_errors(resource)
        return parse_raw_as(Project, resource.text)

    def update_project(
        self, project_id: str, project_input: ProjectInput, request_kwargs: Dict = {}
    ) -> Project:
        response = self._client.patch(
            f"/projects/{project_id}",
            data=project_input.json(exclude_unset=True),
            **request_kwargs,
        )
        handle_errors(response)
        return parse_raw_as(Project, response.text)

    def suggest_project_id(self, display_name: str, request_kwargs: Dict = {}) -> str:
        response = self._client.get(
            "/projects:suggest-id",
            params={"display_name": display_name},
            **request_kwargs,
        )
        handle_errors(response)
        return response.json()

    def delete_project(self, project_id: str, request_kwargs: Dict = {}) -> None:
        response = self._client.delete(f"/projects/{project_id}", **request_kwargs)
        handle_errors(response)

    def list_project_members(
        self, project_id: str, request_kwargs: Dict = {}
    ) -> List[User]:
        response = self._client.get(f"/projects/{project_id}/users", **request_kwargs)
        handle_errors(response)
        return parse_raw_as(List[User], response.text)

    def add_project_member(
        self,
        project_id: str,
        user_id: str,
        access_level: AccessLevel,
        request_kwargs: Dict = {},
    ) -> List[User]:
        resource = self._client.put(
            f"/projects/{project_id}/users/{user_id}",
            params={"access_level": access_level.value},
            **request_kwargs,
        )
        handle_errors(resource)
        return parse_raw_as(List[User], resource.text)

    def remove_project_member(
        self, project_id: str, user_id: str, request_kwargs: Dict = {}
    ) -> List[User]:
        resource = self._client.delete(
            f"/projects/{project_id}/users/{user_id}", **request_kwargs
        )
        handle_errors(resource)
        return parse_raw_as(List[User], resource.text)



from typing import Dict, List, Tuple

from contaxy import config
from contaxy.operations.project import ProjectOperations
from contaxy.schema import Project, User
from contaxy.schema.auth import AccessLevel
from contaxy.schema.exceptions import ClientValueError
from contaxy.schema.extension import GLOBAL_EXTENSION_PROJECT
from contaxy.schema.project import ProjectCreation

PERMISSION_SEPERATOR = "#"
RESOURCE_WILDCARD = "*"  # Allows access to all resources

ACCESS_LEVEL_MAPPING: Dict[AccessLevel, List[AccessLevel]] = {
    AccessLevel.ADMIN: [AccessLevel.WRITE, AccessLevel.READ],
    AccessLevel.WRITE: [AccessLevel.READ],
    AccessLevel.READ: [],
}

ACCESS_LEVEL_REVERSE_MAPPING: Dict[AccessLevel, List[AccessLevel]] = {}
for key, value in ACCESS_LEVEL_MAPPING.items():
    for string in value:
        ACCESS_LEVEL_REVERSE_MAPPING.setdefault(string, []).append(key)


def is_valid_permission(permission_str: str) -> bool:
    """Returns `True` if the `permission_str` is valid permission."""
    # TODO: do some basic checks on the permission string
    return PERMISSION_SEPERATOR in permission_str


def is_jwt_token(token: str) -> bool:
    """Returns `True` if the provided token is an JWT token."""
    # TODO: Improve this detection
    return len(token) != config.settings.API_TOKEN_LENGTH and not token.islower()


def construct_permission(resource_name: str, access_level: AccessLevel) -> str:
    """Constructs a permission based on the provided `resource_name`  and `access_level`."""
    return resource_name + PERMISSION_SEPERATOR + access_level.value


def parse_permission(permission: str) -> Tuple[str, AccessLevel]:
    """Extracts the resource name and access level from a permission.

    Args:
        permission: The permission to parse.

    Raises:
        ClientValueError: If the permission cannot be parsed.

    Returns:
        Tuple[str, AccessLevel]: The extracted resource name and access level.
    """

    if not is_valid_permission(permission):
        raise ClientValueError(f"{permission} is not a valid permission.")

    resource, level_str = permission.split(PERMISSION_SEPERATOR)
    # The resource name should not start or end with a slash or colon
    resource = resource.lstrip("/").lstrip(":").rstrip("/").rstrip(":")
    return resource, AccessLevel.load(level_str)


def is_access_level_granted(
    granted_access_level: AccessLevel, requested_access_level: AccessLevel
) -> bool:
    """Checks if the requested access level is allowed by the granted access level.

    Args:
        granted_access_level: The access level that is already granted.
        requested_access_level: The access level to check against the granted level.

    Returns:
        bool: `True` if the access level is granted.
    """
    if granted_access_level is requested_access_level:
        # Same access level
        return True

    return requested_access_level in ACCESS_LEVEL_MAPPING[granted_access_level]


def is_permission_granted(granted_permission: str, requested_permission: str) -> bool:
    """Checks if the requested permission is allowed by the granted permission.

    Args:
        granted_permission: The permission that is already granted.
        requested_permission: The permission to check against the granted permission

    Raises:
        ClientValueError: If one of the permissions cannot be parsed.

    Returns:
        bool: `True` if the permission is granted.
    """

    granted_perm_resource, granted_perm_level = parse_permission(granted_permission)
    requested_perm_resource, requested_perm_level = parse_permission(
        requested_permission
    )

    # the * is a special -> it allows access to all resources
    # Add path divider (/ and :) since it should only grant access if it matches a full path
    if (
        granted_perm_resource != RESOURCE_WILDCARD
        and not f"{requested_perm_resource}/".startswith(f"{granted_perm_resource}/")
        and not f"{requested_perm_resource}:".startswith(f"{granted_perm_resource}:")
    ):
        # The requested permission does not start with the granted permission
        return False

    if not is_access_level_granted(granted_perm_level, requested_perm_level):
        # The requested permission needs atleast the same level as the granted permission
        return False

    return True


def setup_user(user: User, project_manager: ProjectOperations) -> Project:
    """Execute initial setup required for each new user.

    This includes:
    - Creation of a technical project that belongs to the user, with the same id as the user id.
    - Addition of the new user to the global extension project so the user can access global extensions.

    Args:
        user (User): The user for whom the project shall be created.
        project_manager (ProjectOperations): The project manager instance used to create the project.

    Raises:
        ResourceAlreadyExistsError: if a project with the id of the user already exists.

    Returns:
        Project: The created technical project of which the user is a member.
    """
    if user.username is not None:
        # Split name by space and dot
        user_first_name = user.username.split()[0].split(".")[0]
        user_project_name = f"{user_first_name.capitalize()}'s Home"
    else:
        user_project_name = "Home"
    user_project = project_manager.create_project(
        project_input=ProjectCreation(
            id=user.id,
            display_name=user_project_name,
            description="My personal project nobody else has access to.",
            # TODO: Maybe state that only the admin has access to the project
        ),
        technical_project=True,
    )

    if user_project.id:
        project_manager.add_project_member(user_project.id, user.id, AccessLevel.ADMIN)

    # User requires read access to the global extension project to access the extension endpoints
    project_manager.add_project_member(
        GLOBAL_EXTENSION_PROJECT, user.id, AccessLevel.READ
    )

    return user_project


def parse_userid_from_resource_name(user_resource_name: str) -> str:
    """Returns the user id from a user-resource name.

    For example, 'users/abcd' returns 'abcd' as 'users' is the prefix.

    Args:
        user_resource_name (str): The resource name, e.g. 'users/abcd'.

    Returns:
        str: The user id in the user-resource name or an empty string.
    """

    if not user_resource_name or not user_resource_name.startswith("users/"):
        return ""

    return user_resource_name.split("/")[1]



from abc import ABC, abstractmethod
from typing import List, Optional

from contaxy.schema import Extension, ExtensionInput


class ExtensionOperations(ABC):
    @abstractmethod
    def list_extensions(self, project_id: str) -> List[Extension]:
        pass

    @abstractmethod
    def install_extension(
        self, extension: ExtensionInput, project_id: str
    ) -> Extension:
        pass

    @abstractmethod
    def delete_extension(
        self, project_id: str, extension_id: Optional[str] = None
    ) -> None:
        pass

    @abstractmethod
    def suggest_extension_config(
        self,
        container_image: str,
        project_id: str,
    ) -> ExtensionInput:
        pass

    @abstractmethod
    def get_extension_metadata(
        self,
        project_id: str,
        extension_id: str,
    ) -> Extension:
        pass

    # TODO: v2
    # @abstractmethod
    # def set_extension_defaults(
    #    self,
    #    extension_id: str,
    #    operation_id: List[str],
    #    project_id: str,
    # ) -> Any:
    #    pass

    # TODO: v2
    # @abstractmethod
    # def get_extension_defaults(
    #     self,
    #     project_id: str,
    # ) -> Any:
    #     pass



from abc import ABC, abstractmethod
from typing import List, Optional

from contaxy.schema import (
    AuthorizedAccess,
    OAuth2TokenRequestFormNew,
    OAuthToken,
    OAuthTokenIntrospection,
    TokenType,
    User,
    UserInput,
    UserRegistration,
)
from contaxy.schema.auth import ApiToken


class AuthOperations(ABC):
    @abstractmethod
    def create_token(
        self,
        scopes: List[str],
        token_type: TokenType,
        description: Optional[str] = None,
    ) -> str:
        """Returns a session or API token with access to the speicfied scopes.

        The token is created on behalfe of the authorized user.

        Args:
            scopes: Scopes requested for this token. If none specified, the token will be generated with same set of scopes as the authorized token.
            token_type: The type of the token.
            description (optional): A short description about the generated token.

        Returns:
            str: The created session or API token.
        """
        pass

    @abstractmethod
    def list_api_tokens(self) -> List[ApiToken]:
        """Lists all API tokens associated from the authorized user.

        Returns:
            List[ApiToken]: A list of API tokens.
        """
        pass

    @abstractmethod
    def verify_access(
        self, token: str, permission: Optional[str] = None, disable_cache: bool = False
    ) -> AuthorizedAccess:
        """Verifies if the authorized token is valid and grants a certain permission.

        The token is verfied for its validity and - if provided - if it has the specified permission.

        Args:
            token: Token (session or API) to verify.
            permission (optional): The token is checked if it is granted this permission. If none specified, only the existence or validity of the token itself is checked.
            disable_cache (optional): If `True`, no cache will be used for verifying the token. Defaults to `False`.

        Raises:
            PermissionDeniedError: If the requested permission is denied.
            UnauthenticatedError: If the token is invalid or expired.

        Returns:
            AuthorizedAccess: Information about the granted permission and authenticated user.
        """
        pass

    @abstractmethod
    def change_password(
        self,
        user_id: str,
        password: str,
    ) -> None:
        """Changes the password of a given user.

        The password is stored as a hash.

        Args:
            user_id: The ID of the user.
            password: The new password to apply for the user.
        """
        pass

    # TODO: v2
    # @abstractmethod
    # def authorize_client(form_data: OAuth2AuthorizeRequestForm) -> Any:
    #    pass

    @abstractmethod
    def request_token(
        self, token_request_form: OAuth2TokenRequestFormNew
    ) -> OAuthToken:
        """Returns an access tokens, ID tokens, or refresh tokens depending on the request parameters.

        The token endpoint is used by the client to obtain an access token by
        presenting its authorization grant or refresh token.

        The token endpoint supports the following grant types:
        - [Password Grant](https://tools.ietf.org/html/rfc6749#section-4.3.2): Used when the application exchanges the user’s username and password for an access token.
            - `grant_type` must be set to `password`
            - `username` (required): The user’s username.
            - `password` (required): The user’s password.
            - `scope` (optional): Optional requested scope values for the access token.
        - [Refresh Token Grant](https://tools.ietf.org/html/rfc6749#section-6): Allows to use refresh tokens to obtain new access tokens.
            - `grant_type` must be set to `refresh_token`
            - `refresh_token` (required): The refresh token previously issued to the client.
            - `scope` (optional): Requested scope values for the new access token. Must not include any scope values not originally granted by the resource owner, and if omitted is treated as equal to the originally granted scope.
        - [Client Credentials Grant](https://tools.ietf.org/html/rfc6749#section-4.4.2): Request an access token using only its client
        credentials.
            - `grant_type` must be set to `client_credentials`
            - `scope` (optional): Optional requested scope values for the access token.
            - Client Authentication required (e.g. via client_id and client_secret or auth header)
        - [Authorization Code Grant](https://tools.ietf.org/html/rfc6749#section-4.1): Used to obtain both access tokens and refresh tokens based on an authorization code from the `/authorize` endpoint.
            - `grant_type` must be set to `authorization_code`
            - `code` (required): The authorization code that the client previously received from the authorization server.
            - `redirect_uri` (required): The redirect_uri parameter included in the original authorization request.
            - Client Authentication required (e.g. via client_id and client_secret or auth header)

        For password, client credentials, and refresh token flows, calling this endpoint is the only step of the flow.
        For the authorization code flow, calling this endpoint is the second step of the flow.

        This endpoint implements the [OAuth2 Token Endpoint](https://tools.ietf.org/html/rfc6749#section-3.2).

        Args:
            token_request_form: The request instructions.

        Raises:
            OAuth2Error: If an error occures. Conforms the RFC6749 spec.

        Returns:
            OAuthToken: The access token and additonal metadata (depending on the grant type).
        """
        pass

    @abstractmethod
    def revoke_token(
        self,
        token: str,
        # token_type_hint: Optional[str] = None,
    ) -> None:
        """Revokes a given token.

        This will delete the API token, preventing further requests with the given token.
        Because of caching, the API token might still be usable under certain conditions
        for some operations for a maximum of 15 minutes after deletion.

        This operation implements the OAuth2 Revocation Flow ([RFC7009](https://tools.ietf.org/html/rfc7009)).

        Args:
            token: The token that should be revoked.

        Raises:
            OAuth2Error: If an error occures. Conforms the RFC6749 spec.
        """
        pass

    @abstractmethod
    def introspect_token(
        self,
        token: str,
        # token_type_hint: Optional[str] = None,
    ) -> OAuthTokenIntrospection:
        """Introspects a given token.

        Returns a status (`active`) that indicates whether it is active or not.
        If the token is active, additional data about the token is also returned.
        If the token is invalid, expired, or revoked, it is considered inactive.

        This operation implements the [OAuth2 Introspection Flow](https://www.oauth.com/oauth2-servers/token-introspection-endpoint/) ([RFC7662](https://tools.ietf.org/html/rfc7662)).

        Args:
            token: The token that should be instrospected.

        Returns:
            OAuthTokenIntrospection: The token state and additional metadata.
        """
        pass

    # User Operations

    @abstractmethod
    def list_users(self) -> List[User]:
        """Lists all users.

        TODO: Filter based on authenticated user?

        Returns:
            List[User]: List of users.
        """
        pass

    @abstractmethod
    def create_user(
        self, user_input: UserRegistration, technical_user: bool = False
    ) -> User:
        """Creates a user.

        Args:
            user_input: The user data to create the new user.
            technical_user: If `True`, the created user will be marked as technical user. Defaults to `False`.

        Raises:
            ResourceAlreadyExistsError: If a user with the same username or email already exists.

        Returns:
            User: The created user information.
        """
        pass

    @abstractmethod
    def get_user(self, user_id: str) -> User:
        """Returns the user metadata for a single user.

        Args:
            user_id: The ID of the user.

        Raises:
            ResourceNotFoundError: If no user with the specified ID exists.

        Returns:
            User: The user information.
        """
        pass

    @abstractmethod
    def update_user(self, user_id: str, user_input: UserInput) -> User:
        """Updates the user metadata.

        This will update only the properties that are explicitly set in the `user_input`.
        The patching is based on the JSON Merge Patch Standard [RFC7396](https://tools.ietf.org/html/rfc7396).

        Args:
            user_id (str): The ID of the user.
            user_input (UserInput): The user data used to update the user.

        Raises:
            ResourceNotFoundError: If no user with the specified ID exists.

        Returns:
            User: The updated user information.
        """
        pass

    @abstractmethod
    def delete_user(self, user_id: str) -> None:
        """Deletes a user.

        Args:
            user_id (str): The ID of the user.

        Raises:
            ResourceNotFoundError: If no user with the specified ID exists.
        """
        pass



from datetime import datetime
from typing import List, Literal, Optional

import docker
from loguru import logger

from contaxy.managers.auth import AuthManager
from contaxy.managers.deployment.docker_utils import (
    create_container_config,
    delete_container,
    get_project_container,
    get_project_containers,
    handle_network,
    list_deploy_service_actions,
    map_job,
    map_service,
    read_container_logs,
    reconnect_to_all_networks,
)
from contaxy.managers.deployment.manager import DeploymentManager
from contaxy.managers.deployment.utils import Labels, split_image_name_and_tag
from contaxy.managers.system import SystemManager
from contaxy.schema import Job, JobInput, ResourceAction, Service, ServiceInput
from contaxy.schema.deployment import DeploymentType
from contaxy.schema.exceptions import ClientBaseError, ClientValueError
from contaxy.utils.auth_utils import parse_userid_from_resource_name
from contaxy.utils.state_utils import GlobalState, RequestState


class DockerDeploymentManager(DeploymentManager):
    _is_initialized = False

    def __init__(
        self,
        global_state: GlobalState,
        request_state: RequestState,
        system_manager: SystemManager,
        auth_manager: AuthManager,
    ):
        """Initializes the docker deployment manager.

        Args:
            global_state: The global state of the app instance.
            request_state: The state for the current request.
            system_manager: The system manager used for getting the list of allowed images.
            auth_manager: The auth manager is used to generate api tokens that a passed to services and jobs.
        """
        self.global_state = global_state
        self.request_state = request_state
        self._system_manager = system_manager
        self._auth_manager = auth_manager

        self.client = docker.from_env()
        # Reconnect the backend to all existing docker networks on startup
        if not DockerDeploymentManager._is_initialized:
            reconnect_to_all_networks(self.client)
            DockerDeploymentManager._is_initialized = True

    def list_services(
        self,
        project_id: str,
        deployment_type: Literal[
            DeploymentType.SERVICE, DeploymentType.EXTENSION
        ] = DeploymentType.SERVICE,
    ) -> List[Service]:
        try:
            containers = get_project_containers(
                client=self.client,
                project_id=project_id,
                deployment_type=deployment_type,
            )

            return [map_service(container) for container in containers]
        except docker.errors.APIError:
            return []

    def deploy_service(
        self,
        project_id: str,
        service: ServiceInput,
        action_id: Optional[str] = None,
        deployment_type: Literal[
            DeploymentType.SERVICE, DeploymentType.EXTENSION
        ] = DeploymentType.SERVICE,
    ) -> Service:
        image_name, image_tag = split_image_name_and_tag(service.container_image)
        self._system_manager.check_image(image_name, image_tag)
        container_config = create_container_config(
            service=service,
            project_id=project_id,
            auth_manager=self._auth_manager,
            user_id=parse_userid_from_resource_name(
                self.request_state.authorized_subject
            ),
        )
        container_config["labels"][Labels.DEPLOYMENT_TYPE.value] = deployment_type.value
        handle_network(client=self.client, project_id=project_id)

        try:
            container = self.client.containers.run(**container_config)
        except docker.errors.APIError as e:
            logger.error(f"Error in deploy service '{service.display_name}': {e}")
            raise ClientValueError(
                f"Could not deploy service '{service.display_name}'."
            )

        return map_service(container)

    def list_deploy_service_actions(
        self, project_id: str, service: ServiceInput
    ) -> List[ResourceAction]:
        return list_deploy_service_actions(project_id=project_id, deploy_input=service)

    def get_service_metadata(self, project_id: str, service_id: str) -> Service:
        container = get_project_container(
            client=self.client, project_id=project_id, deployment_id=service_id
        )
        return map_service(container)

    def delete_service(
        self, project_id: str, service_id: str, delete_volumes: bool = False
    ) -> None:
        container = get_project_container(
            self.client, project_id=project_id, deployment_id=service_id
        )
        delete_container(container=container, delete_volumes=delete_volumes)

    def delete_services(
        self,
        project_id: str,
    ) -> None:
        containers = get_project_containers(client=self.client, project_id=project_id)
        for container in containers:
            container.remove(v=True, force=True)

    def get_service_logs(
        self,
        project_id: str,
        service_id: str,
        lines: Optional[int] = None,
        since: Optional[datetime] = None,
    ) -> str:
        container = get_project_container(
            self.client, project_id=project_id, deployment_id=service_id
        )

        return read_container_logs(container=container, lines=lines, since=since)

    def list_jobs(self, project_id: str) -> List[Job]:
        containers = get_project_containers(
            self.client, project_id=project_id, deployment_type=DeploymentType.JOB
        )
        return [map_job(container) for container in containers]

    def deploy_job(
        self,
        project_id: str,
        job: JobInput,
        action_id: Optional[str] = None,
    ) -> Job:
        image_name, image_tag = split_image_name_and_tag(job.container_image)
        self._system_manager.check_image(image_name, image_tag)
        container_config = create_container_config(
            service=job,
            project_id=project_id,
            auth_manager=self._auth_manager,
            user_id=parse_userid_from_resource_name(
                self.request_state.authorized_subject
            ),
        )
        container_config["labels"][
            Labels.DEPLOYMENT_TYPE.value
        ] = DeploymentType.JOB.value
        handle_network(client=self.client, project_id=project_id)

        try:
            container = self.client.containers.run(**container_config)
        except docker.errors.APIError:
            raise ClientBaseError(
                status_code=500, message=f"Could not deploy job '{job.display_name}'."
            )

        response_service = map_job(container)
        return response_service

    def list_deploy_job_actions(
        self,
        project_id: str,
        job: JobInput,
    ) -> List[ResourceAction]:
        return list_deploy_service_actions(project_id=project_id, deploy_input=job)

    def get_job_metadata(self, project_id: str, job_id: str) -> Job:
        container = get_project_container(
            client=self.client,
            project_id=project_id,
            deployment_id=job_id,
            deployment_type=DeploymentType.JOB,
        )

        return map_job(container)

    def delete_job(self, project_id: str, job_id: str) -> None:
        container = get_project_container(
            self.client,
            project_id=project_id,
            deployment_id=job_id,
            deployment_type=DeploymentType.JOB,
        )
        delete_container(container=container)

    def delete_jobs(
        self,
        project_id: str,
    ) -> None:
        containers = get_project_containers(
            self.client, project_id=project_id, deployment_type=DeploymentType.JOB
        )
        for container in containers:
            container.remove(v=True, force=True)

    def get_job_logs(
        self,
        project_id: str,
        job_id: str,
        lines: Optional[int] = None,
        since: Optional[datetime] = None,
    ) -> str:
        container = get_project_container(
            self.client,
            project_id=project_id,
            deployment_id=job_id,
            deployment_type=DeploymentType.JOB,
        )

        return read_container_logs(container=container, lines=lines, since=since)



from abc import ABC, abstractmethod
from typing import List

from pydantic import EmailStr

from contaxy.schema import File, Project, ProjectCreation, User, UserRegistration
from contaxy.utils.file_utils import FileStreamWrapper


class SeedOperations(ABC):
    @abstractmethod
    def create_user(
        self,
        user_input: UserRegistration = UserRegistration(
            username="Foo", email=EmailStr("foo@bar.com"), password="Foobar"
        ),
    ) -> User:
        pass

    @abstractmethod
    def create_users(self, amount: int) -> List[User]:
        pass

    @abstractmethod
    def create_project(
        self,
        project_input: ProjectCreation = ProjectCreation(
            id="my-test-project", display_name="My Test Project!"
        ),
    ) -> Project:
        pass

    @abstractmethod
    def create_projects(self, amount: int) -> List[Project]:
        pass

    @abstractmethod
    def create_file(
        self,
        project_id: str,
        file_key: str = "my-test-file",
        max_number_chars: int = 200,
    ) -> File:
        pass

    @abstractmethod
    def create_files(
        self,
        project_id: str,
        number_of_files: int,
        prefix: str = "my-test-file",
        max_number_chars: int = 200,
    ) -> List[File]:
        pass

    @abstractmethod
    def create_file_stream(
        self,
        max_number_chars: int = 200,
    ) -> FileStreamWrapper:
        pass



from abc import ABC, abstractmethod
from typing import Any, Iterator, List, Optional

from contaxy.schema import File, FileInput, FileStream, ResourceAction


class FileOperations(ABC):
    @abstractmethod
    def list_files(
        self,
        project_id: str,
        recursive: bool = True,
        include_versions: bool = False,
        prefix: Optional[str] = None,
    ) -> List[File]:
        """List files.

        Args:
            project_id (str): Project ID associated with the file.
            recursive (bool, optional): List recursively as directory structure emulation. Defaults to True.
            include_versions (bool, optional): Flag to control whether include object versions. Defaults to False.
            prefix (Optional[str], optional): File key starts with prefix. Defaults to None.

        Returns:
            List[File]: List of file metadata objects.
        """
        pass

    @abstractmethod
    def get_file_metadata(
        self,
        project_id: str,
        file_key: str,
        version: Optional[str] = None,
    ) -> File:
        """Get file metadata of a single file.

        If no version is provided then the latest version will be returned.

        Args:
            project_id (str): Project ID associated with the file.
            file_key (str): Key of the file.
            version (Optional[str], optional): File version. Defaults to None.

        Raises:
            ResourceNotFoundError: If no file is found.

        Returns:
            File: The file metadata object.
        """
        pass

    @abstractmethod
    def update_file_metadata(
        self,
        file: FileInput,
        project_id: str,
        file_key: str,
        version: Optional[str] = None,
    ) -> File:
        """Update the file metadata.

        If no version is provided then the latest version will be returned. Moreover, additional custom metadata provied via `file.metadata` will executed with json merge patch strategy in case a specific version is updated. But if a new version was created it is treated as a new file and hence the metadata from older versions will not be considered. If a version should inherit the metadata from a previous version, then this data needs to be set explictly.

        Args:
            file (FileInput): The file metadata object. All unset attributes or None values will be ignored.
            project_id (str): Project ID associated with the file.
            file_key (str): Key of the file.
            version (Optional[str], optional): File version. Defaults to None.

        Raises:
            ClientValueError: If the provided keys do not match.

        Returns:
            File: The updated file metadata object.
        """
        pass

    @abstractmethod
    def upload_file(
        self,
        project_id: str,
        file_key: str,
        file_stream: FileStream,
        content_type: str = "application/octet-stream",
    ) -> File:
        """Upload a file.

        Args:
            project_id (str): Project ID associated with the file.
            file_key (str): Key of the file.
            file_stream (FileStream): The actual file stream object.
            content_type (str, optional): The mime-type of the file. Defaults to "application/octet-stream".

        Raises:
            ServerBaseError: If the upload failed.

        Returns:
            File: The file metadata object of the uploaded file.
        """
        pass

    @abstractmethod
    def download_file(
        self,
        project_id: str,
        file_key: str,
        version: Optional[str] = None,
    ) -> Iterator[bytes]:
        """Download a file.

        Either the latest version will be returned or the specified one.

        Args:
            project_id (str): Project ID associated with the file.
            file_key (str): Key of the file.
            version (Optional[str], optional): File version. Defaults to None.

        Raises:
            ResourceNotFoundError: If file does not exist.

        Yields:
            Iterator[bytes]: [description]
        """
        pass

    @abstractmethod
    def delete_file(
        self,
        project_id: str,
        file_key: str,
        version: Optional[str] = None,
        keep_latest_version: bool = False,
    ) -> None:
        """Deletes a file.

        If a specific file `version` is provided, only this one will be deleted. If no `version` is provided and `keep_latest_version` is True, all but the latest version will be deleted. Otherwise, all existing versions will be removed.

        Args:
            project_id (str): Project ID associated with the file.
            file_key (str): Key of the file.
            version (Optional[str], optional): File version. Defaults to None.
            keep_latest_version (bool, optional): [description]. Defaults to False.
        """
        pass

    @abstractmethod
    def delete_files(
        self,
        project_id: str,
    ) -> None:
        """Deletes all files associated with a project.

        Args:
            project_id (str): Project ID associated with the file.
        """
        pass

    @abstractmethod
    def list_file_actions(
        self, project_id: str, file_key: str, version: Optional[str] = None
    ) -> List[ResourceAction]:
        pass

    @abstractmethod
    def execute_file_action(
        self,
        project_id: str,
        file_key: str,
        action_id: str,
        version: Optional[str] = None,
    ) -> Any:
        pass



from abc import ABC, abstractmethod, abstractproperty
from typing import List, Optional

from fastapi import Path
from pydantic import BaseModel, ByteSize, Field

from contaxy.schema.shared import Resource, ResourceInput

FILE_KEY_REGEX = r"^(?!.*(\r|\n|\.\.)).{1,1024}$"  # no new lines or consecutive dots

FILE_KEY_PARAM = Path(
    ...,
    title="File Key",
    example="datasets/customer-table.csv",
    description="A valid file key.",
    regex=FILE_KEY_REGEX,
    min_length=1,
    max_length=1024,
)


class FileStream(ABC):
    @abstractproperty
    def hash(self) -> str:
        pass

    @abstractmethod
    def read(self, size: int = -1) -> bytes:
        pass


class FileBase(BaseModel):
    key: str = Field(
        ...,
        example="datasets/customer-table.csv",
        description="The (virtual) path of the file. This path might not correspond to the actual path on the file storage.",
        min_length=1,
        max_length=1024,  # Keys can only be 1024 chars long
        regex=FILE_KEY_REGEX,
    )
    content_type: Optional[str] = Field(
        None,
        example="text/csv",
        description="A standard MIME type describing the format of the contents. If an file is stored without a Content-Type, it is served as application/octet-stream.",
    )
    external_id: Optional[str] = Field(
        None,
        example="datasets/customer-table.csv",
        description="The ID (or access instruction) of the file on the file storage provider.",
    )
    # TODO: v2
    # content_encoding: Specifies what content encodings have been applied to the object and thus what decoding mechanisms must be applied to obtain the media-type referenced by the Content-Type header field.
    # content_disposition: Specifies presentational information for the object.
    # content_language: The language the content is in.


class FileInput(ResourceInput, FileBase):
    pass


class File(Resource, FileBase):
    file_extension: Optional[str] = Field(
        None,
        example="csv",
        description="The full file extension extracted from the key field. May contain multiple concatenated extensions, such as `tar.gz`.",
    )
    file_size: Optional[ByteSize] = Field(
        None,
        example=1073741824,
        description="The file size in bytes.",
    )
    version: Optional[str] = Field(
        None,
        example="1614204512210009",
        description="Version tag of this file. The version order might not be inferable from the version tag.",
    )
    available_versions: Optional[List[str]] = Field(
        None,
        example=["1614204512210009", "23424512210009", "6144204512210009"],
        description="All version tags available for the given file.",
    )
    latest_version: Optional[bool] = Field(
        None,
        example=True,
        description="Indicates if this is the latest available version of the file.",
    )
    md5_hash: Optional[str] = Field(
        None,
        example="2a53375ff139d9837e93a38a279d63e5",
        description="The base64-encoded 128-bit MD5 digest of the file. This can be used for checking the file integrity.",
    )
    etag: Optional[str] = Field(
        None,
        example="57f456164b0e5f365aaf9bb549731f32-95",
        description="The etag of the file (mainly used by S3 storage). An entity-tag is an opaque validator for differentiating between multiple representations of the same resource",
    )
    extension_id: Optional[str] = Field(
        None,
        description="The extension ID in case the file is provided via an extension.",
    )



from enum import Enum
from typing import List, Optional

from fastapi import Query
from pydantic import BaseModel, Field

from contaxy.schema.deployment import Service, ServiceInput
from contaxy.schema.shared import QUALIFIED_RESOURCE_ID_REGEX

CORE_EXTENSION_ID = "core"
GLOBAL_EXTENSION_PROJECT = "ctxy-global"

EXTENSION_ID_PARAM = Query(
    None,
    title="Extension ID",
    description="Extension ID. If not specified, the system will decide. Use `core` to explicitly select the core platform.",
    regex=QUALIFIED_RESOURCE_ID_REGEX,
)


class ExtensionType(str, Enum):
    USER_EXTENSION = "user-extension"
    PROJECT_EXTENSION = "project-extension"


class ExtensionBase(BaseModel):
    capabilities: Optional[List[str]] = Field(
        None,
        # TODO: provide example
        description="List of capabilities implemented by this extension.",
    )
    api_extension_endpoint: Optional[str] = Field(
        None,
        example="8080/extension/api",
        description="The endpoint base URL that implements the API operation stated in the capabilities property.",
    )
    ui_extension_endpoint: Optional[str] = Field(
        None,
        example="8080/webapp/ui",
        description="The endpoint instruction that provide a Web UI. If this is provided, the extension will be integrated into the UI.",
    )
    extension_type: Optional[ExtensionType] = Field(
        None,
        example=ExtensionType.PROJECT_EXTENSION,
        description="The type of the extension.",
    )


class ExtensionInput(ServiceInput, ExtensionBase):
    pass


class Extension(Service, ExtensionBase):
    pass



from enum import Enum
from typing import Dict, List, Optional

from fastapi import Query
from pydantic import BaseModel, Field, validator


class SystemState(str, Enum):
    STARTING = "starting"
    INITIALIZING = "initializing"
    RUNNING = "running"


class SystemInfo(BaseModel):
    version: str = Field(
        ...,
        example="0.1.0",
        description="Platform version.",
    )
    # TODO: orchestrator or other selected settings
    namespace: str = Field(
        ...,
        example="mlhub",
        description="Namespace of this system.",
    )
    system_state: SystemState = Field(
        SystemState.STARTING, description="The state of the system."
    )
    metadata: Optional[Dict[str, str]] = Field(
        None,
        example={"additional-metadata": "value"},
        description="Additional key-value metadata associated with this system.",
    )


class SystemStatistics(BaseModel):
    # TODO: finish model
    project_count: int
    user_count: int
    job_count: int
    service_count: int
    file_count: int


class AllowedImageInfo(BaseModel):
    image_name: str = Field(
        None,
        example="my-docker-registry.com/my-image",
        description="Name of the docker image to allow. Do not specify the image tag (the part after the colon)",
    )
    image_tags: List[str] = Field(
        None,
        example=["0.2.1", "0.3.0"],
        description='List of tags that are allowed for this image. Can be set to ["*"] to allow all tags.',
    )
    metadata: Optional[Dict[str, str]] = Field(
        None,
        example={"additional-metadata": "value"},
        description="A collection of arbitrary key-value pairs associated with this image that does not need predefined structure.",
    )

    @validator("image_name", pre=True)
    def ensure_lowercase(cls, value: str) -> str:
        return value.lower()


IMAGE_NAME_PARAM = Query(
    ...,
    title="Docker image name",
    description="Name of a docker image without the tag",
)



from abc import ABC, abstractmethod
from datetime import datetime
from typing import Any, List, Literal, Optional

from contaxy.schema import Job, JobInput, ResourceAction, Service, ServiceInput
from contaxy.schema.deployment import DeploymentType

# TODO: update_service functionality


class ServiceOperations(ABC):
    @abstractmethod
    def list_services(
        self,
        project_id: str,
        deployment_type: Literal[
            DeploymentType.SERVICE, DeploymentType.EXTENSION
        ] = DeploymentType.SERVICE,
    ) -> List[Service]:
        """Lists all services associated with the given project.

        Args:
            project_id (str): The project ID to filter the services.
            deployment_type (One of [DeploymentType.SERVICE, DeploymentType.JOB]): The deployment type of either Service or Extension (which is a subtype of Service).

        Returns:
            List[Service]: The list of services associated with the project.
        """
        pass

    @abstractmethod
    def deploy_service(
        self,
        project_id: str,
        service: ServiceInput,
        action_id: Optional[str] = None,
        deployment_type: Literal[
            DeploymentType.SERVICE, DeploymentType.EXTENSION
        ] = DeploymentType.SERVICE,
    ) -> Service:
        """Deploys a service for the specified project.

        If no `action_id` is provided, the system will automatically select the best deployment option.

        Available deployment options (actions) can be requested via the [list_deploy_service_actions](#services/list_deploy_service_actions) operation.
        If the action is from an extension, the `action_id` must be a composite ID with the following format: `{extension_id}~{action_id}`.

        The action mechanism is further explained in the description of the [list_deploy_service_actions](#services/list_deploy_service_actions).

        Args:
            project_id (str): [description]
            service (ServiceInput): [description]
            action_id (Optional[str], optional): The ID of the selected action. Defaults to `None`.
            deployment_type (One of [DeploymentType.SERVICE, DeploymentType.JOB]): The deployment type of either Service or Extension (which is a subtype of Service).

        Returns:
            Service: The metadata of the deployed service.
        """
        pass

    @abstractmethod
    def list_deploy_service_actions(
        self,
        project_id: str,
        service: ServiceInput,
    ) -> List[ResourceAction]:
        """Lists all available service deployment options (actions).

        Args:
            project_id (str): The project ID associated with the service.
            service_id (str): The ID of the service.

        Returns:
            List[ResourceAction]: Available deployment actions.
        """
        pass

    @abstractmethod
    def get_service_metadata(
        self,
        project_id: str,
        service_id: str,
    ) -> Service:
        """Returns the metadata of a single service.

        Args:
            project_id (str): The project ID associated with the service.
            service_id (str): The ID of the service.

        Returns:
            Service: The service metadata.
        """
        pass

    @abstractmethod
    def delete_service(
        self,
        project_id: str,
        service_id: str,
        delete_volumes: bool = False,
    ) -> None:
        """Deletes a service.

        Args:
            project_id (str): The project ID associated with the service.
            service_id (str): The ID of the service.
            delete_volumes (bool, optional): If `True`, all attached volumes will be deleted. Defaults to `False`.

        Raises:
            RuntimeError: If an error occurs during the deletion of the service.
        """
        pass

    @abstractmethod
    def delete_services(
        self,
        project_id: str,
    ) -> None:
        """Deletes all services associated with a project.

        Args:
            project_id (str): The project ID.
        """
        pass

    @abstractmethod
    def get_service_logs(
        self,
        project_id: str,
        service_id: str,
        lines: Optional[int],
        since: Optional[datetime],
    ) -> str:
        """Returns the logs of a service.

        Args:
            project_id (str): The ID of the project into which the service is deployed in.
            service_id (str): The ID of the service.
            lines (Optional[int]): If provided, just the last `n` lines are returned from the log. Defaults to `None`.
            since (Optional[datetime]): If provided, just the logs since the given timestamp are returned. Defaults to `None`.

        Raises:
            NotImplementedError: [description]
            RuntimeError: If reading the logs of the given service fails.

        Returns:
            str: The logs of the service.
        """
        pass

    @abstractmethod
    def suggest_service_config(
        self,
        project_id: str,
        container_image: str,
    ) -> ServiceInput:
        """Suggests an input configuration based on the provided `container_image`.

        The suggestion is based on metadata extracted from the container image (e.g. labels)
        as well as suggestions based on previous project deployments with the same image.

        Args:
            project_id (str): The project ID associated with the service.
            container_image (str): The container image to use as context for the suggestion.

        Returns:
            ServiceInput: The suggested service configuration.
        """
        pass

    @abstractmethod
    def list_service_actions(
        self,
        project_id: str,
        service_id: str,
    ) -> List[ResourceAction]:
        """Lists all actions available for the specified service.

        See the endpoint documentation for more information on the action mechanism.

        Args:
            project_id (str): The project ID associated with the service.
            service_id (str): The ID of the service.

        Returns:
            List[ResourceAction]: Available actions for given services.
        """
        pass

    @abstractmethod
    def execute_service_action(
        self,
        project_id: str,
        service_id: str,
        action_id: str,
    ) -> Any:
        """Executes the selected service action.

        The actions need to be first requested from the list_service_actions operation.
        If the action is from an extension, the `action_id` must be a composite ID with the following format: `{extension_id}~{action_id}`.

        Args:
            project_id (str): The project ID associated with the service.
            service_id (str): The ID of the service.
            action_id (str): The ID of the selected action.

        Returns:
            `None` or a redirect response to another URL.
        """
        pass


class JobOperations(ABC):
    @abstractmethod
    def list_jobs(self, project_id: str) -> List[Job]:
        pass

    @abstractmethod
    def deploy_job(
        self,
        project_id: str,
        job: JobInput,
        action_id: Optional[str] = None,
    ) -> Job:
        pass

    @abstractmethod
    def list_deploy_job_actions(
        self,
        project_id: str,
        job: JobInput,
    ) -> List[ResourceAction]:
        pass

    @abstractmethod
    def suggest_job_config(self, project_id: str, container_image: str) -> JobInput:
        pass

    @abstractmethod
    def get_job_metadata(self, project_id: str, job_id: str) -> Job:
        pass

    @abstractmethod
    def delete_job(self, project_id: str, job_id: str) -> None:
        pass

    @abstractmethod
    def delete_jobs(
        self,
        project_id: str,
    ) -> None:
        """Deletes all jobs associated with a project.

        Args:
            project_id (str): The project ID.
        """
        pass

    @abstractmethod
    def get_job_logs(
        self,
        project_id: str,
        job_id: str,
        lines: Optional[int] = None,
        since: Optional[datetime] = None,
    ) -> str:
        pass

    @abstractmethod
    def list_job_actions(
        self,
        project_id: str,
        job_id: str,
    ) -> List[ResourceAction]:
        pass

    @abstractmethod
    def execute_job_action(self, project_id: str, job_id: str, action_id: str) -> Any:
        pass


class DeploymentOperations(ServiceOperations, JobOperations, ABC):
    pass



from typing import Any, List, Optional

from fastapi import APIRouter, Depends, Path, Query, status

from contaxy.api.dependencies import (
    ComponentManager,
    get_api_token,
    get_component_manager,
)
from contaxy.schema import CoreOperations, Extension, ExtensionInput
from contaxy.schema.auth import AccessLevel
from contaxy.schema.exceptions import (
    AUTH_ERROR_RESPONSES,
    CREATE_RESOURCE_RESPONSES,
    GET_RESOURCE_RESPONSES,
    VALIDATION_ERROR_RESPONSE,
)
from contaxy.schema.extension import GLOBAL_EXTENSION_PROJECT
from contaxy.schema.project import PROJECT_ID_PARAM

router = APIRouter(
    tags=["extensions"],
    responses={**AUTH_ERROR_RESPONSES, **VALIDATION_ERROR_RESPONSE},
)


@router.get(
    "/projects/{project_id}/extensions",
    operation_id=CoreOperations.LIST_EXTENSIONS.value,
    response_model=List[Extension],
    summary="List extensions.",
    status_code=status.HTTP_200_OK,
)
def list_extensions(
    project_id: str = PROJECT_ID_PARAM,
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Returns all installed extensions accesible by the specified project.

    This also includes all extensions which are installed globally as well as
    extensions installed by the authorized user.
    """

    component_manager.verify_access(
        token,
        f"/projects/{project_id}/extensions",
        AccessLevel.READ,
    )

    return component_manager.get_extension_manager().list_extensions(
        project_id=project_id
    )


@router.delete(
    "/projects/{project_id}/extensions/{extension_id}",
    operation_id=CoreOperations.DELETE_EXTENSION.value,
    summary="Delete extension.",
    status_code=status.HTTP_204_NO_CONTENT,
)
def delete_extension(
    project_id: str = PROJECT_ID_PARAM,
    extension_id: Optional[str] = Path(
        ...,
        title="Extension ID",
        description="A valid extension ID.",
    ),
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Deletes an extension.

    This will delete the installation metadata as well as the service container.
    """
    raise NotImplementedError


@router.get(
    "/projects/{project_id}/extensions/{extension_id}",
    operation_id=CoreOperations.GET_EXTENSION_METADATA.value,
    response_model=Extension,
    summary="Get extension metadata.",
    status_code=status.HTTP_200_OK,
    responses={**GET_RESOURCE_RESPONSES},
)
def get_extension_metadata(
    project_id: str = PROJECT_ID_PARAM,
    extension_id: str = Path(
        ...,
        title="Extension ID",
        description="A valid extension ID.",
    ),
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Returns the metadata of the given extension."""
    raise NotImplementedError


@router.post(
    "/projects/{project_id}/extensions",
    operation_id=CoreOperations.INSTALL_EXTENSION.value,
    response_model=Extension,
    summary="Install extension.",
    status_code=status.HTTP_200_OK,
    responses={**CREATE_RESOURCE_RESPONSES},
)
def install_extension(
    extension: ExtensionInput,
    project_id: str = PROJECT_ID_PARAM,
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Installs an extension for the given project.

    This will deploy the extension container for the selected project and
    registers the extension for all the specified capabilities.
    """

    # Only admins can install "global" extensions that are shown for every user
    if project_id == GLOBAL_EXTENSION_PROJECT:
        component_manager.verify_access(token, "*", AccessLevel.ADMIN)
    else:
        component_manager.verify_access(
            token,
            f"/projects/{project_id}/extensions",
            AccessLevel.WRITE,
        )

    return component_manager.get_extension_manager().install_extension(
        extension=extension, project_id=project_id
    )


@router.get(
    "/projects/{project_id}/extensions:suggest-config",
    operation_id=CoreOperations.SUGGEST_EXTENSION_CONFIG.value,
    response_model=ExtensionInput,
    summary="Suggest extension configuration.",
    status_code=status.HTTP_200_OK,
)
def suggest_extension_config(
    container_image: str = Query(
        ..., description="Container image to use for suggestion."
    ),
    project_id: str = PROJECT_ID_PARAM,
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Suggests an input configuration based on the provided `container_image`.

    The suggestion is based on metadata extracted from the container image (e.g. labels)
    as well as suggestions based on previous project deployments with the same image.
    """
    raise NotImplementedError



from datetime import datetime
from typing import Optional

from fastapi import Path
from pydantic import BaseModel, Field

from contaxy.schema.shared import Resource, ResourceInput

PROJECTS_KIND = "projects"

MIN_PROJECT_ID_LENGTH = 4
MAX_PROJECT_ID_LENGTH = 25

PROJECT_ID_PARAM = Path(
    ...,
    title="Project ID",
    example="image-search-engine",
    description="A valid project ID.",
    min_length=MIN_PROJECT_ID_LENGTH,
    max_length=MAX_PROJECT_ID_LENGTH,
)


class Quota(BaseModel):
    max_cpus: Optional[int] = Field(
        None,
        example=52,
        ge=0,
        description="Maximum number of CPU cores.",
    )
    max_memory: Optional[int] = Field(
        None,
        example=80000,
        ge=0,
        description="Maximum amount of memory in Megabyte.",
    )
    max_gpus: Optional[int] = Field(
        None,
        example=8,
        ge=0,
        description="Maximum number of GPUs.",
    )
    max_deployment_storage: Optional[int] = Field(
        None,
        example=800000,
        ge=0,
        description="Maximum storage usage in Megabyte for all deployments.",
    )
    max_file_storage: Optional[int] = Field(
        None,
        example=100000,
        ge=0,
        description="Maximum storage usage in Megabyte for all files.",
    )
    max_files: Optional[int] = Field(
        None,
        example=1000,
        ge=0,
        description="Maximum number of files on file storage.",
    )
    max_deployments: Optional[int] = Field(
        None,
        example=20,
        ge=0,
        description="Maximum number of deployments. This includes services, jobs, and extensions.",
    )
    max_collections: Optional[int] = Field(
        None,
        example=20,
        ge=0,
        description="Maximum number of JSON document collections.",
    )


class ProjectBase(BaseModel):
    # TODO: add quota
    # quota: Optional[Quota] = Field(
    #    None,
    #    description="Limitations for resource usage.",
    # )
    pass


class ProjectInput(ResourceInput, ProjectBase):
    pass


class ProjectCreation(ProjectInput):
    id: str = Field(
        ...,
        example="my-awesome-project",
        description="Project ID used for creating the project.",
    )


class Project(Resource, ProjectBase):
    technical_project: Optional[bool] = Field(
        False,
        description="Indicates that this is a technical project created by the system.",
    )
    # TODO: v2
    # statistics: Optional[Statistics] = Field(
    #    None,
    #    description="Project statistics.",
    # )


class Statistics(BaseModel):
    cpus: Optional[int] = Field(
        None,
        ge=0,
        example=32,
        description="Number of CPU cores requested by active deployments.",
    )
    memory: Optional[int] = Field(
        None,
        ge=0,
        example=60000,
        description="Amount of memory in Megabyte requested by active deployments.",
    )
    gpus: Optional[int] = Field(
        None,
        ge=0,
        example=8,
        description="Number of GPUs requested by active deployments.",
    )
    deployment_storage: Optional[int] = Field(
        None,
        ge=0,
        example=400000,
        description="Storage usage in Megabyte for all deployments.",
    )
    file_storage: Optional[int] = Field(
        None,
        ge=0,
        example=50000,
        description="Storage usage in Megabyte for all files.",
    )
    files: Optional[int] = Field(
        None,
        example=320,
        ge=0,
        description="Number of files on file storage.",
    )
    deployments: Optional[int] = Field(
        None,
        example=10,
        ge=0,
        description="Number of deployments. This includes services, jobs, and extensions.",
    )
    collections: Optional[int] = Field(
        None,
        example=10,
        ge=0,
        description="Number of JSON document collections.",
    )
    last_activity: Optional[datetime] = Field(
        None,
        example="2021-04-23T10:20:30.400+02:30",
        description="The last activity based on events.",
    )



from typing import List, Optional, Tuple, Union

from contaxy import config
from contaxy.clients import DeploymentManagerClient, FileClient
from contaxy.clients.shared import BaseUrlSession
from contaxy.managers.deployment.manager import DeploymentManager
from contaxy.managers.deployment.utils import (
    Labels,
    get_template_mapping,
    replace_template_string,
)
from contaxy.operations import ExtensionOperations
from contaxy.schema import ExtensibleOperations, Extension, ExtensionInput, Service
from contaxy.schema.deployment import DeploymentType, ServiceInput
from contaxy.schema.extension import (
    CORE_EXTENSION_ID,
    GLOBAL_EXTENSION_PROJECT,
    ExtensionType,
)
from contaxy.utils.auth_utils import parse_userid_from_resource_name
from contaxy.utils.state_utils import GlobalState, RequestState

COMPOSITE_ID_SEPERATOR = "~"

CAPABILITIES_METADATA_SEPARATOR = ","
METADATA_UI_EXTENSION_ENDPOINT = "ctxy.ui_extension_endpoint"
METADATA_API_EXTENSION_ENDPOINT = "ctxy.api_extension_endpoint"
METADATA_CAPABILITIES = "ctxy.capabilities"
METADATA_EXTENSION_TYPE = "ctxy.extensionType"


def parse_composite_id(composite_id: str) -> Tuple[str, Union[str, None]]:
    """Extracts the resource and extension ID from an composite ID.

    If the provided ID does not contain an composite ID seperator (`~`)
    the ID will be returned as resource ID and the extension ID will be `None`.

    Args:
        composite_id (str): The ID to parse.

    Returns:
        Tuple[str, str]: Resource ID, Extension ID
    """
    # TODO: Only sperate based on the first occurance

    if COMPOSITE_ID_SEPERATOR not in composite_id:
        # The provided id does not contain an extension ID prefix
        return composite_id, None

    resource_id, extension_id = composite_id.split(COMPOSITE_ID_SEPERATOR, 1)
    return resource_id, extension_id


def map_service_to_extension(service: Service, user_id: str) -> Extension:
    extension = Extension(**service.dict())

    if service.metadata:
        project_id = service.metadata[Labels.PROJECT_NAME.value]
        deployment_id = service.metadata[Labels.DEPLOYMENT_NAME.value]
        endpoint_prefix = f"{config.settings.CONTAXY_BASE_URL}/projects/{project_id}/services/{deployment_id}/access/"

        if METADATA_CAPABILITIES in service.metadata:
            extension.capabilities = service.metadata[METADATA_CAPABILITIES].split(
                CAPABILITIES_METADATA_SEPARATOR
            )

        template_mapping = get_template_mapping(
            project_id=project_id,
            user_id=parse_userid_from_resource_name(user_id),
        )

        if METADATA_UI_EXTENSION_ENDPOINT in service.metadata:
            service_ui_extension_endpoint = replace_template_string(
                input=service.metadata[METADATA_UI_EXTENSION_ENDPOINT],
                templates_mapping=template_mapping,
            )
            extension.ui_extension_endpoint = (
                f"{endpoint_prefix}{service_ui_extension_endpoint}"
            )

        if METADATA_API_EXTENSION_ENDPOINT in service.metadata:
            service_api_extension_endpoint = replace_template_string(
                input=service.metadata[METADATA_API_EXTENSION_ENDPOINT],
                templates_mapping=template_mapping,
            )
            extension.api_extension_endpoint = (
                f"{endpoint_prefix}{service_api_extension_endpoint}"
            )

        if METADATA_EXTENSION_TYPE in service.metadata:
            try:
                extension.extension_type = ExtensionType(
                    service.metadata[METADATA_EXTENSION_TYPE]
                )
            except ValueError:
                pass

    return extension


class ExtensionClient(FileClient, DeploymentManagerClient):
    """Handels the request forwarding to the installed extensions.

    The extension client implements all extensible manager interfaces
    and uses the generated HTTP client to forward requests to the respective extension.

    Extension clients are initialized and managed by the `ExtensionManager`.
    """

    def __init__(self, endpoint_url: str, access_token: Optional[str] = None):
        """Initializes the extension client.

        Args:
            endpoint_url: The endpoint base URL of the extension.
            access_token: An access token that is attached to all requests.
        """
        client_session = BaseUrlSession(base_url=endpoint_url)
        if access_token:
            client_session.headers.update({config.API_TOKEN_NAME: access_token})
            # client_session.cookies.set(config.API_TOKEN_NAME, access_token)
        super(ExtensionClient, self).__init__(client=client_session)


class ExtensionManager(ExtensionOperations):
    """Installs and manages extensions.

    The extension manager implements all methods for the full extension lifecycle.
    It is the central entry point for all interactions with extensions.
    """

    def __init__(
        self,
        global_state: GlobalState,
        request_state: RequestState,
        deployment_manager: DeploymentManager,
    ):
        """Initializes the extension manager.

        Args:
            global_state: The global state of the app instance.
            request_state: The state for the current request.
            deployment_manager: The current deployment manager instance.
        """
        self.global_state = global_state
        self.request_state = request_state
        self.deployment_manager = deployment_manager

    def get_extension_client(self, extension_id: str) -> ExtensionClient:
        endpoint_url = f"http://{config.settings.SYSTEM_NAMESPACE}-{extension_id}:8080"
        access_token = (
            self.request_state.authorized_access.access_token.token
            if self.request_state.authorized_access
            and self.request_state.authorized_access.access_token
            else ""
        )
        return ExtensionClient(
            endpoint_url=endpoint_url,
            access_token=access_token,
        )

    def list_extensions(self, project_id: str) -> List[Extension]:
        if project_id != GLOBAL_EXTENSION_PROJECT:
            project_services = self.deployment_manager.list_services(
                project_id=project_id, deployment_type=DeploymentType.EXTENSION
            )
        else:
            project_services = []
        global_services = self.deployment_manager.list_services(
            project_id=GLOBAL_EXTENSION_PROJECT,
            deployment_type=DeploymentType.EXTENSION,
        )

        extension_services = []
        for service in [*project_services, *global_services]:
            # TODO: Add a function to deployment-manager that allows direct filtering
            if service.deployment_type != DeploymentType.EXTENSION.value:
                continue

            extension = map_service_to_extension(
                service=service,
                user_id=self.request_state.authorized_subject,
            )
            extension_services.append(extension)

        return extension_services

    def get_default_extension(self, operation: ExtensibleOperations) -> str:
        return CORE_EXTENSION_ID

    def delete_extension(
        self, project_id: str, extension_id: Optional[str] = None
    ) -> None:
        raise NotImplementedError

    def get_extension_metadata(self, project_id: str, extension_id: str) -> Extension:
        raise NotImplementedError

    def install_extension(
        self, extension: ExtensionInput, project_id: str
    ) -> Extension:
        service_input = ServiceInput(**extension.dict())
        if not service_input.metadata:
            service_input.metadata = {}
        if extension.capabilities:
            service_input.metadata[
                METADATA_CAPABILITIES
            ] = CAPABILITIES_METADATA_SEPARATOR.join(extension.capabilities)
        if extension.ui_extension_endpoint:
            service_input.metadata[
                METADATA_UI_EXTENSION_ENDPOINT
            ] = extension.ui_extension_endpoint
        if extension.api_extension_endpoint:
            service_input.metadata[
                METADATA_API_EXTENSION_ENDPOINT
            ] = extension.api_extension_endpoint
        if extension.extension_type:
            service_input.metadata[METADATA_EXTENSION_TYPE] = extension.extension_type

        service = self.deployment_manager.deploy_service(
            project_id=project_id,
            service=service_input,
            deployment_type=DeploymentType.EXTENSION,
        )
        return map_service_to_extension(
            service=service,
            user_id=self.request_state.authorized_subject,
        )

    def suggest_extension_config(
        self, container_image: str, project_id: str
    ) -> ExtensionInput:
        raise NotImplementedError



import time
from datetime import datetime, timezone
from typing import List, Literal, Optional

from kubernetes import client as kube_client
from kubernetes import config as kube_config
from kubernetes.client.models import (
    V1Deployment,
    V1DeploymentList,
    V1Job,
    V1JobList,
    V1JobSpec,
    V1ServiceList,
    V1Status,
)
from kubernetes.client.rest import ApiException
from loguru import logger

from contaxy.config import settings
from contaxy.managers.auth import AuthManager
from contaxy.managers.deployment.kube_utils import (
    build_deployment_metadata,
    build_kube_deployment_config,
    build_kube_service_config,
    build_pod_template_spec,
    build_project_network_policy_spec,
    check_or_create_project_network_policy,
    create_pvc,
    create_service,
    get_deployment_selection_labels,
    get_label_selector,
    get_pod,
    map_kube_job,
    map_kube_service,
    wait_for_deletion,
    wait_for_deployment,
    wait_for_job,
)
from contaxy.managers.deployment.manager import DeploymentManager
from contaxy.managers.deployment.utils import (
    DEFAULT_DEPLOYMENT_ACTION_ID,
    NO_LOGS_MESSAGE,
    Labels,
    get_deployment_id,
    split_image_name_and_tag,
)
from contaxy.managers.system import SystemManager
from contaxy.schema import Job, JobInput, ResourceAction, Service, ServiceInput
from contaxy.schema.deployment import DeploymentType
from contaxy.schema.exceptions import (
    ClientBaseError,
    ClientValueError,
    ResourceNotFoundError,
    ServerBaseError,
)
from contaxy.utils.auth_utils import parse_userid_from_resource_name
from contaxy.utils.state_utils import GlobalState, RequestState


class KubernetesDeploymentManager(DeploymentManager):
    def __init__(
        self,
        global_state: GlobalState,
        request_state: RequestState,
        system_manager: SystemManager,
        auth_manager: AuthManager,
        kube_namespace: str = None,
    ):
        """Initializes the Kubernetes Deployment Manager.

        Args:
            global_state: The global state of the app instance.
            request_state: The state for the current request.
            system_manager: The system manager used for getting the list of allowed images.
            auth_manager: The auth manager is used to generate api tokens that a passed to services and jobs.
            kube_namespace (str): Set the Kubernetes namespace to use. If it is not given, the manager will try to detect the namespace automatically.
        """
        self.global_state = global_state
        self.request_state = request_state
        self._system_manager = system_manager
        self._auth_manager = auth_manager

        try:
            # incluster config is the config given by a service account and it's role permissions
            kube_config.load_incluster_config()
        except kube_config.ConfigException:
            kube_config.load_kube_config()

        self.core_api = kube_client.CoreV1Api()
        self.apps_api = kube_client.AppsV1Api()
        self.batch_api = kube_client.BatchV1Api()
        self.networking_api = kube_client.NetworkingV1Api()

        if kube_namespace is None:
            try:
                # at this path the namespace the container is in is stored in Kubernetes deployment (see https://stackoverflow.com/questions/31557932/how-to-get-the-namespace-from-inside-a-pod-in-openshift)
                with open(
                    "/var/run/secrets/kubernetes.io/serviceaccount/namespace", "r"
                ) as namespace_file:
                    self.kube_namespace = namespace_file.read()
            except FileNotFoundError:
                # TODO: fix arguments
                raise ServerBaseError("Could not detect the Kubernetes Namespace")
        else:
            self.kube_namespace = kube_namespace
        # TODO: when we have performance problems in the future, replicate the watch logic from JupyterHub KubeSpawner to keep Pod & other resource information in memory? (see https://github.com/jupyterhub/kubespawner/blob/941585f0f7acb0f366c9979b6274b7f47356a630/kubespawner/reflector.py#L238)

    def list_services(
        self,
        project_id: str,
        deployment_type: Literal[
            DeploymentType.SERVICE, DeploymentType.EXTENSION
        ] = DeploymentType.SERVICE,
    ) -> List[Service]:
        label_selector = get_deployment_selection_labels(
            project_id=project_id, deployment_type=deployment_type
        )

        try:
            deployments: V1DeploymentList = self.apps_api.list_namespaced_deployment(
                namespace=self.kube_namespace, label_selector=label_selector
            )
            return [map_kube_service(deployment) for deployment in deployments.items]
        except ApiException:
            return []

    def deploy_service(
        self,
        project_id: str,
        service: ServiceInput,
        action_id: Optional[str] = None,
        deployment_type: Literal[
            DeploymentType.SERVICE, DeploymentType.EXTENSION
        ] = DeploymentType.SERVICE,
        wait: bool = False,
    ) -> Service:
        image_name, image_tag = split_image_name_and_tag(service.container_image)
        self._system_manager.check_image(image_name, image_tag)
        if service.display_name is None:
            raise ClientValueError(
                message=f"Could not create a service id for service with display name {service.display_name}",
                explanation="A display name for the service must be provided.",
            )

        service_id = get_deployment_id(
            project_id=project_id,
            deployment_name=service.display_name,
            deployment_type=deployment_type,
        )

        kube_service_config = build_kube_service_config(
            service_id=service_id,
            service=service,
            project_id=project_id,
            kube_namespace=self.kube_namespace,
        )
        kube_deployment_config, kube_deployment_pvc = build_kube_deployment_config(
            service_id=service_id,
            service=service,
            project_id=project_id,
            kube_namespace=self.kube_namespace,
            auth_manager=self._auth_manager,
            user_id=parse_userid_from_resource_name(
                self.request_state.authorized_subject
            ),
        )

        check_or_create_project_network_policy(
            network_policy=build_project_network_policy_spec(
                project_id=project_id, kube_namespace=self.kube_namespace
            ),
            networking_api=self.networking_api,
        )

        create_pvc(
            pvc=kube_deployment_pvc,
            kube_namespace=self.kube_namespace,
            core_api=self.core_api,
        )
        create_service(
            service_config=kube_service_config,
            kube_namespace=self.kube_namespace,
            core_api=self.core_api,
        )

        try:
            deployment: V1Deployment = self.apps_api.create_namespaced_deployment(
                namespace=self.kube_namespace, body=kube_deployment_config
            )

            if wait:
                wait_for_deployment(
                    deployment_name=deployment.metadata.name,
                    kube_namespace=self.kube_namespace,
                    apps_api=self.apps_api,
                )

        except (ApiException, Exception) as e:
            # Delete service again as the belonging deployment could not be created, but only when status code is not 409 as 409 indicates that the deployment already exists
            if not hasattr(e, "status") or e.status != 409:  # type: ignore
                try:
                    self.core_api.delete_namespaced_service(
                        namespace=self.kube_namespace, name=service_id
                    )
                except ApiException:
                    pass

                # TODO: delete pvc here

            raise ClientBaseError(
                status_code=500,
                message=f"Could not create namespaced deployment '{service.display_name}' with reason: {e}",
            )

        try:
            transformed_service = map_kube_service(deployment)
        except Exception as e:
            # Delete already created resources upon an error
            try:
                self.core_api.delete_namespaced_service(
                    namespace=self.kube_namespace, name=service_id
                )
            except ApiException:
                pass

            try:
                self.apps_api.delete_namespaced_deployment(
                    namespace=self.kube_namespace, name=service_id
                )
            except ApiException:
                pass

            # TODO: delete pvc here

            raise ServerBaseError(
                f"Could not transform deployment '{service.display_name}' with reason: {e}"
            )

        return transformed_service

    def list_deploy_service_actions(
        self, project_id: str, service: ServiceInput
    ) -> List[ResourceAction]:
        # TODO: make some cluster checks?
        return [
            ResourceAction(
                action_id=DEFAULT_DEPLOYMENT_ACTION_ID,
                display_name=DEFAULT_DEPLOYMENT_ACTION_ID,
            )
        ]

    def get_service_metadata(self, project_id: str, service_id: str) -> Service:
        try:
            deployment: V1Deployment = self.apps_api.read_namespaced_deployment(
                name=service_id, namespace=self.kube_namespace
            )

            # Make sure that the service belongs to the same contaxy namespace as the core-backend. Also, double check that this service really belongs to the project (even though the serviceId should be unique)
            if (
                deployment.metadata.labels[Labels.NAMESPACE.value]
                != settings.SYSTEM_NAMESPACE
                or deployment.metadata.labels[Labels.PROJECT_NAME.value] != project_id
            ):
                raise ResourceNotFoundError(
                    f"Could not get metadata of service '{service_id}' for project {project_id}."
                )

            return map_kube_service(deployment)
        except ApiException:
            raise ResourceNotFoundError(
                f"Could not get metadata of service '{service_id}' for project {project_id}."
            )

    def delete_service(
        self,
        project_id: str,
        service_id: str,
        delete_volumes: bool = False,
        retries: int = 0,
    ) -> None:

        try:
            status: V1Status = self.core_api.delete_namespaced_service(
                name=service_id,
                namespace=self.kube_namespace,
                propagation_policy="Foreground",
            )
            if status.status == "Failure":
                raise ServerBaseError(
                    f"Could not delete Kubernetes service for service-id {service_id}"
                )

            status = self.apps_api.delete_namespaced_deployment(
                name=service_id,
                namespace=self.kube_namespace,
                propagation_policy="Foreground",
            )
            if status.status == "Failure":
                raise ServerBaseError(
                    f"Could not delete Kubernetes deployment for service-id {service_id}"
                )

            if delete_volumes:
                status = self.core_api.delete_namespaced_persistent_volume_claim(
                    namespace=self.kube_namespace, name=service_id
                )
                if status.status == "Failure":
                    # TODO: if we work with a queue system, then add it to a deletion queue
                    # log(
                    #     f"Could not delete Kubernetes Persistent Volume Claim for service-id {service_id}"
                    # )
                    raise ServerBaseError(
                        f"Could not delete Kubernetes Persistent Volume Claim for service-id {service_id}"
                    )

            # wait some time for the deployment to be deleted
            wait_for_deletion(
                self.apps_api, self.kube_namespace, deployment_id=service_id
            )
        except Exception:
            # TODO: add resources to delete to a queue instead of deleting directly? This would have the advantage that even if an operation failes, it is repeated. Also, if the delete endpoint is called multiple times, it is only added once to the queue
            # if retries < max_retries:
            #     try:
            #         return self.delete_service(
            #             project_id=project_id,
            #             service_id=service_id,
            #             delete_volumes=delete_volumes,
            #             retries=retries + 1,
            #         )
            #     except Exception:
            #         pass
            raise ClientBaseError(
                status_code=500,
                message=f"Could not delete service '{service_id}'.",
            )

    def delete_services(
        self,
        project_id: str,
    ) -> None:
        label_selector = get_deployment_selection_labels(
            project_id=project_id, deployment_type=DeploymentType.SERVICE
        )

        try:
            services: V1ServiceList = self.core_api.list_namespaced_service(
                namespace=self.kube_namespace, label_selector=label_selector
            )

            for service in services.items:
                self.core_api.delete_namespaced_service(
                    name=service.metadata.name, namespace=self.kube_namespace
                )

            status = self.apps_api.delete_collection_namespaced_deployment(
                namespace=self.kube_namespace,
                label_selector=label_selector,
                propagation_policy="Foreground",
            )

            if status.status == "Failure":
                raise ServerBaseError(
                    f"Could not delete Kubernetes deployments for project '{project_id}'"
                )

            status = self.core_api.delete_collection_namespaced_persistent_volume_claim(
                namespace=self.kube_namespace,
                label_selector=label_selector,
                propagation_policy="Foreground",
            )

            if status.status == "Failure":
                raise ServerBaseError(
                    f"Could not delete Kubernetes volumes for project '{project_id}'"
                )

        except Exception as e:
            logger.error(f"Error in Kubernetes->delete_services. Reason: {e}")
            raise ClientBaseError(
                500, f"Could not delete services for project '{project_id}'"
            )

    def get_service_logs(
        self,
        project_id: str,
        service_id: str,
        lines: Optional[int] = None,
        since: Optional[datetime] = None,
    ) -> str:
        try:
            pod = get_pod(
                project_id=project_id,
                service_id=service_id,
                kube_namespace=self.kube_namespace,
                core_api=self.core_api,
            )
        except Exception:
            pod = None

        if pod is None:
            raise ResourceNotFoundError(
                f"Could not find service {service_id} to read logs from."
            )

        # TODO: remove as this should not be a concern of the get_logs function
        try:
            # Give some time to let the container within the pod start
            start = time.time()
            timeout = 60
            while True:
                if pod.status.phase in ["Pending", "ContainerCreating"]:
                    try:
                        pod = get_pod(
                            project_id=project_id,
                            service_id=service_id,
                            kube_namespace=self.kube_namespace,
                            core_api=self.core_api,
                        )
                        time.sleep(1)
                    except Exception:
                        pass
                else:
                    break

                if time.time() - start > timeout:
                    raise ServerBaseError(
                        f"Could not read logs from service {service_id} due to status error."
                    )

            since_seconds = None
            if since:
                since_seconds = (
                    int((datetime.now(timezone.utc) - since).total_seconds()) + 1
                )

            try:
                return self.core_api.read_namespaced_pod_log(
                    name=pod.metadata.name,
                    namespace=self.kube_namespace,
                    pretty="true",
                    tail_lines=lines if lines else None,
                    since_seconds=since_seconds,
                )
            except ApiException:
                raise ServerBaseError(f"Could not read logs of service {service_id}.")
        except Exception:
            return NO_LOGS_MESSAGE

    def list_jobs(self, project_id: str) -> List[Job]:
        label_selector = get_label_selector(
            [
                (Labels.NAMESPACE.value, settings.SYSTEM_NAMESPACE),
                (Labels.PROJECT_NAME.value, project_id),
                (Labels.DEPLOYMENT_TYPE.value, DeploymentType.JOB.value),
            ]
        )

        try:
            jobs: V1JobList = self.batch_api.list_namespaced_job(
                namespace=self.kube_namespace, label_selector=label_selector
            )
        except ApiException:
            return []

        return [map_kube_job(job) for job in jobs.items]

    def deploy_job(
        self,
        project_id: str,
        job: JobInput,
        action_id: Optional[str] = None,
        wait: bool = False,
    ) -> Job:
        image_name, image_tag = split_image_name_and_tag(job.container_image)
        self._system_manager.check_image(image_name, image_tag)
        if job.display_name is None:
            raise ClientValueError(
                message=f"Could not create service id for job {job.display_name}",
                explanation="The display name for a service must be set.",
            )

        deployment_id = get_deployment_id(
            project_id=project_id,
            deployment_name=job.display_name,
            deployment_type=DeploymentType.JOB,
        )
        metadata = build_deployment_metadata(
            kube_namespace=self.kube_namespace,
            project_id=project_id,
            deployment_id=deployment_id,
            display_name=job.display_name.replace(" ", "__"),
            labels=job.metadata,
            compute_resources=job.compute,
            endpoints=job.endpoints,
            deployment_type=DeploymentType.JOB,
            user_id=parse_userid_from_resource_name(
                self.request_state.authorized_subject
            ),
        )

        # For debugging purposes, set restart_policy=Never to have access to job logs (see https://kubernetes.io/docs/concepts/workloads/controllers/job/#pod-backoff-failure-policy)
        pod_spec = build_pod_template_spec(
            project_id=project_id,
            service_id=deployment_id,
            service=job,
            metadata=metadata,
            auth_manager=self._auth_manager,
            user_id=parse_userid_from_resource_name(
                self.request_state.authorized_subject
            ),
        )
        pod_spec.spec.restart_policy = "OnFailure"

        _job = V1Job(metadata=metadata, spec=V1JobSpec(template=pod_spec))
        try:
            deployed_job = self.batch_api.create_namespaced_job(
                namespace=self.kube_namespace, body=_job
            )
        except ApiException:
            raise ClientBaseError(
                status_code=500, message=f"Could not deploy job '{job.display_name}'."
            )

        if wait:
            wait_for_job(
                job_name=deployed_job.metadata.name,
                kube_namespace=self.kube_namespace,
                batch_api=self.batch_api,
            )

        return map_kube_job(job=deployed_job)

    def list_deploy_job_actions(
        self,
        project_id: str,
        job: JobInput,
    ) -> List[ResourceAction]:
        # TODO: make some cluster checks?
        return [
            ResourceAction(
                action_id=DEFAULT_DEPLOYMENT_ACTION_ID,
                display_name=DEFAULT_DEPLOYMENT_ACTION_ID,
            )
        ]

    def get_job_metadata(self, project_id: str, job_id: str) -> Job:
        try:
            job: V1Job = self.batch_api.read_namespaced_job(
                name=job_id, namespace=self.kube_namespace
            )
            return map_kube_job(job)
        except ApiException:
            raise ResourceNotFoundError(f"Could not get metadata of job '{job_id}'.")

    def delete_job(self, project_id: str, job_id: str) -> None:
        try:
            self.batch_api.delete_namespaced_job(
                name=job_id,
                namespace=self.kube_namespace,
                propagation_policy="Foreground",
            )

            # wait some time for the job to be deleted
            wait_for_deletion(self.batch_api, self.kube_namespace, job_id)
        except ApiException as e:
            raise ClientBaseError(
                status_code=500,
                message=f"Could not delete job '{job_id}' with reason: {e}",
            )

    def delete_jobs(
        self,
        project_id: str,
    ) -> None:
        label_selector = get_deployment_selection_labels(
            project_id=project_id, deployment_type=DeploymentType.JOB
        )

        try:
            status: V1Status = self.batch_api.delete_collection_namespaced_job(
                namespace=self.kube_namespace,
                label_selector=label_selector,
                propagation_policy="Foreground",
            )

            if status.status == "Failure":
                raise ServerBaseError(
                    f"Could not delete Kubernetes jobs for project '{project_id}'"
                )
        except ApiException as e:
            log = f"Could not delete Kubernetes jobs for project '{project_id}'"
            logger.error(f"{log}. Reason: {e}")
            raise ServerBaseError(log)

    def get_job_logs(
        self,
        project_id: str,
        job_id: str,
        lines: Optional[int] = None,
        since: Optional[datetime] = None,
    ) -> str:
        return self.get_service_logs(
            project_id=project_id, service_id=job_id, lines=lines, since=since
        )



#!/usr/bin/env python

import os
import re
from glob import glob
from os.path import basename, splitext

from setuptools import find_packages, setup  # type: ignore

NAME = "contaxy"
MAIN_PACKAGE = NAME  # Change if main package != NAME
DESCRIPTION = "Python package template."
URL = "https://github.com/ml-tooling/contaxy"
EMAIL = "team@mltooling.org"
AUTHOR = "ML Tooling Team"
LICENSE = "MIT"
REQUIRES_PYTHON = ">=3.8"
VERSION = None  # Only set version if you like to overwrite the version in _about.py

PWD = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
with open(os.path.join(PWD, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

# Extract the version from the _about.py module.
if not VERSION:
    with open(os.path.join(PWD, "src", MAIN_PACKAGE, "_about.py")) as f:  # type: ignore
        VERSION = re.findall(r"__version__\s*=\s*\"(.+)\"", f.read())[0]

# Where the magic happens:
setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    license=LICENSE,
    packages=find_packages(where="src", exclude=("tests", "test", "examples", "docs")),
    package_dir={"": "src"},
    py_modules=[splitext(basename(path))[0] for path in glob("src/*.py")],
    zip_safe=False,
    install_requires=[
        # Add all the runtime requirements here:
        "typer",  # TODO: remove typer?
        "pydantic",
        "fastapi",
        "kubernetes",
        "docker",
        "psutil",
        "uvicorn",
        "addict",
        "sqlalchemy>=1.4.0",
        # Postgres Driver
        "psycopg2",
        # Generates concise, unambiguous, URL-safe UUIDs.
        "shortuuid",
        # Create slugs from unicode strings
        "python-slugify",
        # Better logging
        "loguru",
        # Required by fastapi.security OAuth2PasswordBearer & fastapi.UploadFile for example
        "python-multipart",
        # Used for multipart stream parsing in file manager
        "streaming_form_data",
        # Used in MinioFileManager
        "filetype",
        "minio",
        # Used in AzureBlobFileManager
        "azure-storage-blob",
        # Used for jwt handling
        "python-jose[cryptography]",
        # Used for password hashing
        "passlib[bcrypt]",
        # TODO: FOR in-memory dict db: Merge dictionaries via json merge patch
        "json-merge-patch",
        # TODO: FOR in-memory dict db: Merge dictionaries via json merge patch
        "jsonpath-ng",
        # Email validation with pydantic
        "email-validator",
        # TODO: GraphQL support - remove
        # "graphene",
        # TODO: Dev only - timing
        "fastapi-utils",
        # TODO: Added because of an missing import error
        "pydantic[dotenv]",
        # Needed for admin registration HTML Form w/o the webapp
        # TODO: Improve
        "jinja2",
        # Used for OIDC handling
        "requests_oauthlib",
    ],
    # deprecated: dependency_links=dependency_links,
    extras_require={
        # Add all extras (e.g. for build and test) here:
        # extras can be installed via: pip install package[dev]
        "dev": [
            "setuptools",
            "wheel",
            "twine",
            "flake8",
            "pytest",
            "pytest-mock",
            "pytest-cov",
            "mypy",
            "types-python-slugify",
            "types-requests",
            "types-cachetools",
            "black",
            "pydocstyle",
            "isort",
            "lazydocs",
            "locust",
            # Test profiling
            "pyinstrument",
            # Export profiling information about the tests
            "pytest-profiling",
            # Create fake data for testing
            "faker",
            # For better print debugging via debug
            "devtools[pygments]",
            # For Jupyter Kernel support
            "ipykernel",
            # TODO: Move to required when necessary
            "universal-build",
            "requests",
        ],
    },
    include_package_data=True,
    package_data={
        # If there are data files included in your packages that need to be
        # 'sample': ['package_data.dat'],
        "contaxy.api.endpoints": ["templates/*"]
    },
    classifiers=[
        # TODO: Update based on https://pypi.org/classifiers/
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Information Technology",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Software Development",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering",
        "Topic :: Utilities",
    ],
    project_urls={
        "Changelog": URL + "/releases",
        "Issue Tracker": URL + "/issues",
        "Documentation": URL + "#documentation",
        "Source": URL,
    },
    entry_points={"console_scripts": [f"{NAME}={MAIN_PACKAGE}._cli:cli"]},
    keywords=[
        # eg: 'keyword1', 'keyword2', 'keyword3',
    ],
)



from typing import List, Optional

from contaxy import __version__, config
from contaxy.config import settings
from contaxy.managers.auth import AuthManager
from contaxy.operations import SystemOperations
from contaxy.operations.json_db import JsonDocumentOperations
from contaxy.operations.project import ProjectOperations
from contaxy.schema import (
    ClientValueError,
    ResourceAlreadyExistsError,
    ResourceNotFoundError,
)
from contaxy.schema.auth import ADMIN_ROLE, USERS_KIND, AccessLevel, UserRegistration
from contaxy.schema.project import ProjectCreation
from contaxy.schema.system import AllowedImageInfo, SystemInfo, SystemStatistics
from contaxy.utils import auth_utils
from contaxy.utils.state_utils import GlobalState, RequestState


class SystemManager(SystemOperations):
    _SYSTEM_INFO_COLLECTION = "system-info"
    _ALLOWED_IMAGES_COLLECTION = "allowed-images"

    def __init__(
        self,
        global_state: GlobalState,
        request_state: RequestState,
        json_db_manager: JsonDocumentOperations,
        auth_manager: AuthManager,
        project_manager: ProjectOperations,
    ):
        """Initializes the system manager.

        Args:
            global_state: The global state of the app instance.
            request_state: The state for the current request.
            json_db_manager: Json document manager instance.
            auth_manager: Auth manager instance.
            project_manager: Project manager instance.
        """
        self._global_state = global_state
        self._request_state = request_state
        self._auth_manager = auth_manager
        self._project_manager = project_manager
        self._json_db_manager = json_db_manager

    def get_system_info(self) -> SystemInfo:
        return SystemInfo(
            version=__version__,
            namespace=settings.SYSTEM_NAMESPACE,
        )

    def is_healthy(self) -> bool:
        # TODO: do real healthchecks
        return True

    def get_system_statistics(self) -> SystemStatistics:
        # TODO: Implement system statistics
        return SystemStatistics(
            project_count=0, user_count=0, job_count=0, service_count=0, file_count=0
        )

    def initialize_system(
        self,
        password: Optional[str] = None,
    ) -> None:
        # Don't execute initialization if there are already existing users
        # TODO: This does not prevent the usage of the contaxy API before the system is initialized
        if len(self._auth_manager.list_users()) > 0:
            raise ResourceAlreadyExistsError("The system has already been initialized")

        # Remove authorized access info
        self._request_state.authorized_access = None

        # TODO: do not clean up?
        # Clean up all documents
        for project in self._project_manager.list_projects():
            assert project.id is not None
            self._json_db_manager.delete_json_collections(project.id)

        # Create Admin Role
        self._auth_manager.add_permission(
            ADMIN_ROLE,
            auth_utils.construct_permission("*", AccessLevel.ADMIN),
        )

        # Create Admin User
        # Set initial user password -> SHOULD be changed after the first login
        admin_user = self._auth_manager.create_user(
            UserRegistration(
                username=config.SYSTEM_ADMIN_USERNAME,
                password=password or config.SYSTEM_ADMIN_INITIAL_PASSWORD,  # type: ignore
            ),
            technical_user=True,
        )

        # Add admin role to admin user
        # TODO: use resource name
        admin_user_resource_name = USERS_KIND + "/" + admin_user.id
        self._auth_manager.add_permission(
            admin_user_resource_name,
            ADMIN_ROLE,
        )

        # Create System Internal Project
        self._project_manager.create_project(
            ProjectCreation(
                id=config.SYSTEM_INTERNAL_PROJECT,
                display_name=config.SYSTEM_INTERNAL_PROJECT,
                description="This project holds all system internal services and data.",
            ),
            technical_project=True,
        )

        auth_utils.setup_user(admin_user, self._project_manager)

    def check_image(self, image_name: str, image_tag: str) -> None:
        # If allowed image list is empty (default), then allow all images
        if len(self.list_allowed_images()) == 0:
            return
        try:
            allowed_image_info = self.get_allowed_image(image_name)
        except ResourceNotFoundError:
            raise ClientValueError(
                f"Image {image_name} is not on the list of allowed images!"
            )

        if not (
            "*" in allowed_image_info.image_tags
            or image_tag in allowed_image_info.image_tags
        ):
            raise ClientValueError(
                f"Image {image_name} is on the list of allowed images but tag {image_tag} is not allowed!"
            )

    def add_allowed_image(self, allowed_image: AllowedImageInfo) -> AllowedImageInfo:
        allowed_image_doc = self._json_db_manager.create_json_document(
            config.SYSTEM_INTERNAL_PROJECT,
            self._ALLOWED_IMAGES_COLLECTION,
            key=allowed_image.image_name,
            json_document=allowed_image.json(),
            upsert=True,
        )
        return AllowedImageInfo.parse_raw(allowed_image_doc.json_value)

    def list_allowed_images(self) -> List[AllowedImageInfo]:
        allowed_image_docs = self._json_db_manager.list_json_documents(
            config.SYSTEM_INTERNAL_PROJECT,
            self._ALLOWED_IMAGES_COLLECTION,
        )
        return [
            AllowedImageInfo.parse_raw(doc.json_value) for doc in allowed_image_docs
        ]

    def get_allowed_image(self, image_name: str) -> AllowedImageInfo:
        allowed_image_doc = self._json_db_manager.get_json_document(
            config.SYSTEM_INTERNAL_PROJECT,
            self._ALLOWED_IMAGES_COLLECTION,
            key=image_name,
        )
        return AllowedImageInfo.parse_raw(allowed_image_doc.json_value)

    def delete_allowed_image(self, image_name: str) -> None:
        self._json_db_manager.delete_json_document(
            config.SYSTEM_INTERNAL_PROJECT,
            self._ALLOWED_IMAGES_COLLECTION,
            key=image_name,
        )



import os
from typing import Any, List

from fastapi import APIRouter, Body, Depends, Form, Response, status
from fastapi.templating import Jinja2Templates
from loguru import logger
from starlette.requests import Request
from starlette.responses import HTMLResponse

from contaxy import config
from contaxy.api.dependencies import (
    ComponentManager,
    get_api_token,
    get_component_manager,
)
from contaxy.schema import CoreOperations, SystemInfo, SystemStatistics
from contaxy.schema.auth import AccessLevel
from contaxy.schema.exceptions import (
    AUTH_ERROR_RESPONSES,
    VALIDATION_ERROR_RESPONSE,
    ClientValueError,
)
from contaxy.schema.system import IMAGE_NAME_PARAM, AllowedImageInfo

HERE = os.path.abspath(os.path.dirname(__file__))

templates = Jinja2Templates(directory=os.path.join(HERE, "templates"))

router = APIRouter(
    tags=["system"],
    responses={**VALIDATION_ERROR_RESPONSE},
)


@router.get(
    "/system/info",
    operation_id=CoreOperations.GET_SYSTEM_INFO.value,
    response_model=SystemInfo,
    summary="Get system info.",
    status_code=status.HTTP_200_OK,
)
def get_system_info(
    component_manager: ComponentManager = Depends(get_component_manager),
) -> Any:
    """Returns information about this instance."""
    return component_manager.get_system_manager().get_system_info()


@router.get(
    "/system/health",
    status_code=status.HTTP_204_NO_CONTENT,
    summary="Check server health status.",
)
def check_health(
    component_manager: ComponentManager = Depends(get_component_manager),
) -> Any:
    """Returns a successful return code if the instance is healthy."""
    if component_manager.get_system_manager().is_healthy():
        return Response(status_code=status.HTTP_204_NO_CONTENT)
    else:
        return Response(status_code=status.HTTP_503_SERVICE_UNAVAILABLE)


@router.get(
    "/system/statistics",
    operation_id=CoreOperations.GET_SYSTEM_STATISTICS.value,
    response_model=SystemStatistics,
    summary="Get system statistics.",
    status_code=status.HTTP_200_OK,
    responses={**AUTH_ERROR_RESPONSES},
)
def get_system_statistics(
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    """Returns statistics about this instance."""
    component_manager.verify_access(token, "system", AccessLevel.READ)
    return component_manager.get_system_manager().get_system_statistics()


@router.post(
    "/system/initialize",
    operation_id=CoreOperations.INITIALIZE_SYSTEM.value,
    summary="Initialize the system.",
    status_code=status.HTTP_204_NO_CONTENT,
)
def initialize_system(
    component_manager: ComponentManager = Depends(get_component_manager),
) -> Any:
    """Initializes the system."""
    component_manager.get_system_manager().initialize_system()
    return Response(status_code=status.HTTP_204_NO_CONTENT)


@router.get(
    "/system/admin-form",
    include_in_schema=False,
    response_class=HTMLResponse,
    status_code=status.HTTP_200_OK,
)
def display_register_admin_form(
    request: Request,
    component_manager: ComponentManager = Depends(get_component_manager),
) -> Any:
    if component_manager.get_auth_manager().list_users():
        logger.error("Admin user already exsists")
        return Response(status_code=status.HTTP_409_CONFLICT)

    route = os.path.join(
        config.settings.CONTAXY_BASE_URL,
        config.settings.CONTAXY_API_PATH,
        "system/admin",
    )
    if not route.startswith("/"):
        route = f"/{route}"

    # TODO: use request.url_for in template when settings.CONTAXY_API_PATH is set as root path in fastapi app
    return templates.TemplateResponse(
        "register-admin.html.j2",
        {
            "request": request,
            "route": route,
            "username": config.SYSTEM_ADMIN_USERNAME,
        },
    )


@router.post("/system/admin", status_code=status.HTTP_200_OK)
def register_admin_user(
    request: Request,
    password: str = Form(...),
    password_confirm: str = Form(...),
    component_manager: ComponentManager = Depends(get_component_manager),
) -> Any:
    if component_manager.get_auth_manager().list_users():
        logger.error("Admin user already exsists")
        return Response(status_code=status.HTTP_409_CONFLICT)
    if password != password_confirm:
        return ClientValueError("The passwords do not match.")

    component_manager.get_system_manager().initialize_system(password)

    return templates.TemplateResponse(
        "register-admin-success.html.j2",
        {
            "request": request,
            "username": config.SYSTEM_ADMIN_USERNAME,
        },
    )


@router.post(
    "/system/allowed-images",
    operation_id=CoreOperations.ADD_ALLOWED_IMAGE.value,
    response_model=AllowedImageInfo,
    summary="Add an image to the list of allowed images or replace one already on the list.",
    status_code=status.HTTP_200_OK,
)
def add_allowed_image(
    allowed_image_input: AllowedImageInfo = Body(...),
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    component_manager.verify_access(token, "system/allowed-images", AccessLevel.WRITE)
    return component_manager.get_system_manager().add_allowed_image(allowed_image_input)


@router.get(
    "/system/allowed-images",
    operation_id=CoreOperations.LIST_ALLOWED_IMAGES.value,
    response_model=List[AllowedImageInfo],
    summary="List all allowed images.",
    status_code=status.HTTP_200_OK,
)
def list_allowed_images(
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    component_manager.verify_access(token, "system/allowed-images", AccessLevel.READ)
    return component_manager.get_system_manager().list_allowed_images()


@router.delete(
    "/system/allowed-images",
    operation_id=CoreOperations.DELETE_ALLOWED_IMAGE.value,
    summary="Remove an image from the list of allowed images.",
    status_code=status.HTTP_204_NO_CONTENT,
)
def delete_allowed_image(
    image_name: str = IMAGE_NAME_PARAM,
    component_manager: ComponentManager = Depends(get_component_manager),
    token: str = Depends(get_api_token),
) -> Any:
    component_manager.verify_access(token, "system/allowed-images", AccessLevel.WRITE)
    component_manager.get_system_manager().delete_allowed_image(image_name)
    return Response(status_code=status.HTTP_204_NO_CONTENT)



from abc import ABC, abstractmethod
from typing import List, Optional

from contaxy.schema import JsonDocument


class JsonDocumentOperations(ABC):
    @abstractmethod
    def create_json_document(
        self,
        project_id: str,
        collection_id: str,
        key: str,
        json_document: str,
        upsert: bool = True,
    ) -> JsonDocument:
        """Creates a JSON document for a given key.

        If a document already exists for the given key, the document will be overwritten if `upsert` is True, otherwise an error is raised.

        Args:
            project_id: Project ID associated with the collection.
            collection_id: ID of the collection (database) to use to store this JSON document.
            key: Key of the JSON document.
            json_document: The actual JSON document value.
            upsert: If `True`, the document will be updated/overwritten if it already exists.

        Raises:
            ClientValueError: If the given json_document does not contain valid json.
            ResourceAlreadyExistsError: If a document already exists for the given key and `upsert` is False.

        Returns:
            JsonDocument: The created JSON document.
        """
        pass

    @abstractmethod
    def update_json_document(
        self,
        project_id: str,
        collection_id: str,
        key: str,
        json_document: str,
    ) -> JsonDocument:
        """Updates a JSON document.

        The update is applied on the existing document based on the JSON Merge Patch Standard [RFC7396](https://tools.ietf.org/html/rfc7396).

        Args:
            project_id: Project ID associated with the collection.
            collection_id: ID of the collection (database) that the JSON document is stored in.
            key: Key of the JSON document.
            json_document: The actual JSON document value.

        Raises:
            ResourceNotFoundError: If no JSON document is found with the given `key`.

        Returns:
            JsonDocument: The updated JSON document.
        """
        pass

    @abstractmethod
    def list_json_documents(
        self,
        project_id: str,
        collection_id: str,
        filter: Optional[str] = None,
        keys: Optional[List[str]] = None,
    ) -> List[JsonDocument]:
        """Lists all JSON documents for the given project collection.

        Args:
            project_id: Project ID associated with the collection.
            collection_id: ID of the collection (database) that the JSON document is stored in.
            filter (optional): Allows to filter the result documents based on a JSONPath expression ([JSON Path Specification](https://goessner.net/articles/JsonPath/)). The filter is only applied to filter documents in the list. It is not usable to extract specific properties.
            keys (Optional[List[str]], optional): Json Document Ids, i.e. DB row keys. Defaults to None.

        Raises:
            ClientValueError: If filter is provided and does not contain a valid Json Path filter.

        Returns:
            List[JsonDocument]: List of JSON documents.
        """
        pass

    @abstractmethod
    def get_json_document(
        self,
        project_id: str,
        collection_id: str,
        key: str,
    ) -> JsonDocument:
        """Returns a single JSON document.

        Args:
            project_id: Project ID associated with the JSON document.
            collection_id: ID of the collection (database) that the JSON document is stored in.
            key: Key of the JSON document.

        Raises:
            ResourceNotFoundError: If no JSON document is found with the given `key`.

        Returns:
            JsonDocument: A JSON document.
        """
        pass

    @abstractmethod
    def delete_json_document(
        self,
        project_id: str,
        collection_id: str,
        key: str,
    ) -> None:
        """Deletes a single JSON document.

        If no other document exists in the project collection, the collection will be deleted.

        Args:
            project_id: Project ID associated with the JSON document.
            collection_id: ID of the collection (database) that the JSON document is stored in.
            key: Key of the JSON document.

        Raises:
            ResourceNotFoundError: If no JSON document is found with the given `key`.
        """
        pass

    @abstractmethod
    def delete_json_collections(
        self,
        project_id: str,
    ) -> None:
        """Deletes all JSON collections for a project.

        Args:
            project_id: Project ID associated with the collections.
        """
        pass

    @abstractmethod
    def delete_json_collection(
        self,
        project_id: str,
        collection_id: str,
    ) -> None:
        """Deletes all documents of a single JSON collection.

        Args:
            project_id: Project ID associated with the collection.
            collection_id: ID of the JSON collection (database).
        """
        pass



from abc import ABC, abstractmethod
from typing import List

from contaxy.schema import AccessLevel, Project, ProjectCreation, ProjectInput, User


class ProjectOperations(ABC):
    @abstractmethod
    def list_projects(self) -> List[Project]:
        """Returns all projects visible to the authenticated user."""
        pass

    @abstractmethod
    def create_project(
        self, project_input: ProjectCreation, technical_project: bool = False
    ) -> Project:
        pass

    @abstractmethod
    def get_project(self, project_id: str) -> Project:
        pass

    @abstractmethod
    def update_project(self, project_id: str, project_input: ProjectInput) -> Project:
        pass

    @abstractmethod
    def suggest_project_id(self, display_name: str) -> str:
        pass

    @abstractmethod
    def delete_project(self, project_id: str) -> None:
        pass

    @abstractmethod
    def list_project_members(self, project_id: str) -> List[User]:
        pass

    @abstractmethod
    def add_project_member(
        self,
        project_id: str,
        user_id: str,
        access_level: AccessLevel,
    ) -> List[User]:
        pass

    @abstractmethod
    def remove_project_member(self, project_id: str, user_id: str) -> List[User]:
        pass
